from abc import ABCMeta, abstractmethod


class OptimizerABC(metaclass=ABCMeta):
    """
    Abstract base class for optimizers used in Search CV.

    This class provides a framework for optimizing estimators using parameters assigned by Search CV, such as GridSearchCV, RandomizedSearchCV, etc.

    Parameters:
    *args: Variable length argument list.
    **kwargs: Arbitrary keyword arguments.

    Methods:
    --------
    prepare_data():
        Prepare the data for optimization.

    optimize(*args, **kwargs):
        Optimize the estimator using the assigned parameters.

    get_best_estimator(*args, **kwargs):
        Return the best_estimator, a product of Search CV.

    get_optimized_object(*args, **kwargs):
        Return the entire optimized object, a product of Search CV.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the OptimizerABC class.

        Parameters:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
        """
        pass

    @abstractmethod
    def prepare_data(self):
        """
        Prepare the data for optimization.
        """
        pass

    @abstractmethod
    def optimize(self, *args, **kwargs):
        """
        Optimize the estimator using the assigned parameters.

        Parameters:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
        """
        pass

    @abstractmethod
    def get_best_estimator(self, *args, **kwargs):
        """
        Return the best_estimator, a product of Search CV.

        Parameters:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
        """
        pass

    @abstractmethod
    def get_optimized_object(self, *args, **kwargs):
        """
        Return the entire optimized object, a product of Search CV.

        Parameters:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
        """
        pass



# Importing essential numerical libraries
import numpy as np  # For numerical operations

# Importing optimization library
import optuna  # For hyperparameter optimization
import pandas as pd  # For data manipulation and analysis

# Importing CatBoost libraries
from catboost import *  # For CatBoost algorithms

# Importing Imbalanced-learn library for handling imbalanced datasets
from imblearn.ensemble import (
    BalancedRandomForestClassifier,
)  # For balancing classes in random forests

# Importing InterpretML for model explainability
from interpret.blackbox import *  # For blackbox model explainability
from interpret.glassbox import *  # For glassbox model explainability

# Importing LightGBM libraries
from lightgbm import *  # For LightGBM algorithms

# Importing Optuna integrations
from optuna.integration import OptunaSearchCV  # For hyperparameter search with Optuna

# Importing Ray libraries for distributed computing
from ray import air, tune  # For parallel and distributed computing
from ray.air import session  # For managing ray sessions

# Importing Ray Tune for Scikit-learn
from ray.tune.sklearn import (  # For hyperparameter tuning with Ray
    TuneGridSearchCV,
    TuneSearchCV,
)

# Importing scikit-learn libraries for various machine learning algorithms and utilities
from sklearn.ensemble import *  # Ensemble methods like Random Forest, AdaBoost, etc.
from sklearn.linear_model import *  # Linear models like Logistic Regression, Ridge, etc.
from sklearn.metrics import *  # For model evaluation
from sklearn.model_selection import (  # For model selection utilities
    GridSearchCV,
    RandomizedSearchCV,
    cross_val_score,
    train_test_split,
)
from sklearn.neural_network import *  # For neural network models
from sklearn.svm import *  # For Support Vector Machines

# Importing XGBoost and XGBSE libraries
from xgboost import *  # For XGBoost algorithms
from xgbse import *  # Assuming this is for some specific XGBoost operations

# Importing custom utility and logging libraries
from lohrasb import logger  # Custom logging utility
from lohrasb.abstracts.optimizers import (
    OptimizerABC,
)  # Custom abstract class for optimizers
from lohrasb.decorators.decorators import trackcalls  # Custom decorators
from lohrasb.utils.helper_funcs import _trail_params_retrive  # Custom helper function
from lohrasb.utils.metrics import *  # Custom metrics for evaluation


class OptunaSearch(OptimizerABC):

    """
    OptunaSearch class for optimizing an estimator using Optuna.

    Parameters:
    -----------
    X : pd.DataFrame or numpy 2D array
        Input for training.

    y : pd.DataFrame or numpy 1D array
        Target or label.

    args : tuple
        Args of the class.

    kwargs : dict
        Kwargs of the class.

        study_search_kwargs : dict
            All arguments for creating a study using Optuna. Read more from `study.create` for Optuna at [Optuna Documentation](https://optuna.readthedocs.io/en/stable/reference/generated/optuna.create_study.html).

        main_optuna_kwargs : dict
            Some required arguments for creating `BaseModel`:

            - estimator : object
                An unfitted estimator that has `fit` and `predict` methods.

            - estimator_params : dict
                Parameters that were passed to find the best estimator using the optimization method.

            - measure_of_accuracy : str
                Measurement of performance for classification and regression estimator during hyperparameter optimization while estimating the best estimator.

        optimize_kwargs : dict
            All arguments for the `optimize` object of Optuna, except `objective`. Check at [Optuna Study.optimize](https://optuna.readthedocs.io/en/stable/reference/generated/optuna.study.Study.html#optuna.study.Study.optimize).

        train_test_split_kwargs : dict
            All arguments for `train_test_split` function, such as `cv`, `random_state`, except the *array. Check at [scikit-learn train_test_split](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html).

        fit_optuna_kwargs : dict
            All additional parameters to the `fit()` method of the estimator during training, except `X` and `y`.

    Returns:
    --------
    None

    Methods:
    --------
    prepare_data():
        Prepare data to be consumed by TuneSearchCV. Pass for TuneSearchCV case.

    optimize():
        Optimize estimator using Optuna engine.

    get_optimized_object():
        Get the grid search cv after invoking fit.

    get_best_estimator():
        Return the best estimator if already fitted.

    Notes:
    ------
    It is recommended to use available factories to create a new instance of this class.
    """

    def __init__(
        self,
        X,
        y,
        *args,
        **kwargs,
    ):
        self.study_search_kwargs = kwargs["kwargs"].get("study_search_kwargs", {})
        self.main_optuna_kwargs = kwargs["kwargs"].get("main_optuna_kwargs", {})
        self.optimize_kwargs = kwargs["kwargs"].get("optimize_kwargs", {})
        self.train_test_split_kwargs = kwargs["kwargs"].get(
            "train_test_split_kwargs", {}
        )
        self.fit_optuna_kwargs = kwargs["kwargs"].get("fit_optuna_kwargs", {})
        self.__optuna_search = None
        self.__best_estimator = None
        self.X = X
        self.y = y
        self.X_train = None
        self.y_train = None
        self.X_test = None
        self.y_test = None
        self.args = args

    @property
    def study_search_kwargs(self):
        return self._study_search_kwargs

    @study_search_kwargs.setter
    def study_search_kwargs(self, value):
        self._study_search_kwargs = value

    @property
    def main_optuna_kwargs(self):
        return self._main_optuna_kwargs

    @main_optuna_kwargs.setter
    def main_optuna_kwargs(self, value):
        self._main_optuna_kwargs = value

    @property
    def optimize_kwargs(self):
        return self._optimize_kwargs

    @optimize_kwargs.setter
    def optimize_kwargs(self, value):
        self._optimize_kwargs = value

    @property
    def train_test_split_kwargs(self):
        return self._train_test_split_kwargs

    @train_test_split_kwargs.setter
    def train_test_split_kwargs(self, value):
        self._train_test_split_kwargs = value

    @property
    def fit_optuna_kwargs(self):
        return self._fit_optuna_kwargs

    @fit_optuna_kwargs.setter
    def fit_optuna_kwargs(self, value):
        self._fit_optuna_kwargs = value

    def prepare_data(self):
        """
        Prepare data to be consumed by the optimizer.
        """
        # Split the data into training and testing sets using the train_test_split function
        # The resulting splits are assigned to self.X_train, self.X_test, self.y_train, and self.y_test
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            self.X, self.y, **self.train_test_split_kwargs
        )

        return self

    def optimize(self):
        """
        Optimize estimator using Optuna engine.
        """

        def objective(trial):
            # Calculate the metric for evaluation using CalcMetrics
            calc_metric = CalcMetrics(
                y_true=self.y_test,
                y_pred=None,
                metric=self.main_optuna_kwargs["measure_of_accuracy"],
            )
            # Create a metric calculator using calc_make_scorer method
            metric_calculator = calc_metric.calc_make_scorer(
                self.main_optuna_kwargs["measure_of_accuracy"],
            )
            # Retrieve estimator parameters and fit parameters
            estimator_params = self.main_optuna_kwargs["estimator_params"]
            fit_params = self.fit_optuna_kwargs
            estimator = self.main_optuna_kwargs["estimator"]

            # Retrieve parameters from the trial and create an estimator
            params = _trail_params_retrive(trial, estimator_params)
            if fit_params is not None:
                est = eval(
                    estimator.__class__.__name__
                    + "(**params)"
                    + ".fit(self.X_train, self.y_train, **fit_params)"
                )
            else:
                est = eval(
                    estimator.__class__.__name__
                    + "(**params)"
                    + ".fit(self.X_train, self.y_train)"
                )

            # Make predictions using the estimator
            y_pred = est.predict(self.X_test)

            # Calculate the evaluation metric
            if (
                metric_calculator.__class__.__name__ == "_BaseScorer"
                or metric_calculator.__class__.__name__ == "_ProbaScorer"
                or metric_calculator.__class__.__name__ == "_PredictScorer"
                or metric_calculator.__class__.__name__ == "_ThresholdScorer"
            ):
                raise TypeError(
                    "make_scorer is not supported for Optuna optimizer! Read examples and documentations."
                )
            func_str = metric_calculator
            accr = eval(func_str)
            return accr

        # Create an Optuna study
        study = optuna.create_study(**self.study_search_kwargs)

        # Optimize the study using the objective function
        study.optimize(objective, **self.optimize_kwargs)

        # Handle refit and create the final estimator
        for key, value in self.main_optuna_kwargs.items():
            if key == "refit":
                if value:
                    flag = True
                    logger.info(
                        "If refit is set to True, the optimal model will be refit on the entire dataset, i.e., X_train and y_train!"
                    )
                else:
                    flag = False

        if self.fit_optuna_kwargs != {}:
            if flag:
                est = eval(
                    self.main_optuna_kwargs["estimator"].__class__.__name__
                    + "(**study.best_trial.params)"
                    + ".fit(self.X, self.y, **self.fit_optuna_kwargs)"
                )
            else:
                est = eval(
                    self.main_optuna_kwargs["estimator"].__class__.__name__
                    + "(**study.best_trial.params)"
                    + ".fit(self.X_train, self.y_train, **self.fit_optuna_kwargs)"
                )
        else:
            if flag:
                est = eval(
                    self.main_optuna_kwargs["estimator"].__class__.__name__
                    + "(**study.best_trial.params)"
                    + ".fit(self.X, self.y)"
                )
            else:
                est = eval(
                    self.main_optuna_kwargs["estimator"].__class__.__name__
                    + "(**study.best_trial.params)"
                    + ".fit(self.X_train, self.y_train)"
                )

        # Store the best estimator and the Optuna search results
        self.__best_estimator = est
        self.__optuna_search = study.best_trial
        return self

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the model to the training data.

        If the best estimator is already available, return self.
        Otherwise, optimize the estimator using the `optimize` method.

        Parameters:
        -----------
        X : pd.DataFrame or numpy 2D array
            Input for training.
        y : pd.DataFrame or numpy 1D array
            Target or label.
        args : tuple
            Additional positional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns:
        --------
        self : object
            Returns self if the best estimator is already available, or after optimizing the estimator.
        """

        if not self.__best_estimator:
            self.optimize()

        return self

    def get_optimized_object(self):
        """
        Get the best trial from the Optuna study.

        Returns:
        --------
        best_trial : optuna.trial.Trial
            The best trial found during the optimization process.

        """
        return self.__optuna_search

    def get_best_estimator(self):
        """
        Get the best estimator after invoking the fit method.

        Returns:
        --------
        best_estimator : object
            The best estimator found during the optimization process.

        Raises:
        -------
        NotImplementedError:
            If the best estimator is not available (i.e., it is None), indicating that the fit method has not been implemented.

        """
        if not self.__best_estimator:
            raise NotImplementedError(
                "It seems the best estimator is None. Maybe the fit method is not implemented!"
            )

        return self.__best_estimator


class GridSearch(OptimizerABC):
    """
    Class Factories for initializing BestModel optimizing engines, i.e.,
    GridSearchCV.

    Parameters
    ----------

        X : pd.DataFrame
            Input for training
        y : pd.DataFrame
            Target or label

        args : tuple
            Args of the class.

        kwargs : dict
            Kwargs of the class.

            grid_search_kwargs : dict

            All arguments for creating a study using GridSearch. Read at [GridSeachCV Documentation](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html).
                For example:
                - estimator : object
                    An unfitted estimator that has `fit` and `predict` methods.

                - param_grid : dict
                    Parameters that were passed to find the best estimator using the optimization method.

                - measure_of_accuracy : str
                    Measurement of performance for classification and regression estimator during hyperparameter optimization while estimating the best estimator.

            main_grid_kwargs : dict
                Some required arguments for creating `BaseModel`:

            fit_grid_kwargs : dict
                All additional parameters to the `fit()` method of the estimator during training, except `X` and `y`.

    Return
    ----------

    The best estimator of estimator optimized by GridSearchCV.

    Methods
    -------
    prepare_data()
        Prepare data to be consumed by GridSearchCV.Pass for GridSearchCV case.
    optimize()
        Optimize estimator using GridSearchCV engine.
    get_optimized_object()
        Get the grid search cv  after invoking fit.
    get_best_estimator()
        Return the best estimator if already fitted.
    Notes
    -----
    It is recommended to use available factories
    to create a new instance of this class.

    """

    def __init__(
        self,
        X,
        y,
        *args,
        **kwargs,
    ):
        self.grid_search_kwargs = kwargs["kwargs"].get("grid_search_kwargs", {})
        self.main_grid_kwargs = kwargs["kwargs"].get("main_grid_kwargs", {})
        self.fit_grid_kwargs = kwargs["kwargs"].get("fit_grid_kwargs", {})
        self.__grid_search = None
        self.__best_estimator = None
        self.__X = X
        self.__y = y
        self.args = args

    @property
    def grid_search_kwargs(self):
        return self._grid_search_kwargs

    @grid_search_kwargs.setter
    def grid_search_kwargs(self, value):
        self._grid_search_kwargs = value

    @property
    def main_grid_kwargs(self):
        return self._main_grid_kwargs

    @main_grid_kwargs.setter
    def main_grid_kwargs(self, value):
        self._main_grid_kwargs = value

    @property
    def fit_grid_kwargs(self):
        return self._fit_grid_kwargs

    @fit_grid_kwargs.setter
    def fit_grid_kwargs(self, value):
        self._fit_grid_kwargs = value

    def prepare_data(self):
        """
        Prepare data to be consumed by GridSearchCV.
        """
        pass

    @trackcalls
    def optimize(self):
        """
        Optimize estimator using GridSearchCV engine.
        """
        self.__grid_search = GridSearchCV(**self.grid_search_kwargs)
        self.__grid_search.fit(self.__X, self.__y, *self.args, **self.fit_grid_kwargs)
        self.__best_estimator = self.__grid_search.best_estimator_

        return self

    def fit(self, X, y, *args, **kwargs):
        for key, value in self.grid_search_kwargs.items():
            if key == "refit":
                if value:
                    self.__grid_search.fit(
                        self.__X, self.__y, *self.args, **self.fit_grid_kwargs
                    )
                    logger.info(
                        "If refit is set to True, the optimal model will be refit on the entire dataset again!"
                    )
        self.__best_estimator = self.__grid_search.best_estimator_
        return self

    def get_best_estimator(self, *args, **kwargs):
        """
        Get the best estimator after invoking fit on it.
        """
        if self.__best_estimator is not None:
            return self.__best_estimator
        else:
            logger.error("The best estimator is None !")

    def get_optimized_object(self, *args, **kwargs):
        """
        Get the grid search cv  after invoking fit.
        """
        if self.__grid_search is not None:
            return self.__grid_search
        else:
            raise NotImplementedError(
                "GridSearch has not been implemented \
                or best_estomator is null"
            )


class NewOptunaSearch(OptimizerABC):

    """
    Class for initializing BestModel optimizing engines, i.e., GridSearchCV.

    This class provides factories for initializing the BestModel optimizing engine, OptunaSearchCV.
    It allows for preparing data, optimizing an estimator, and accessing the optimized object and the best estimator.

    Parameters
    ----------
    X : pd.DataFrame
        Input for training.
    y : pd.DataFrame
        Target or label.
    *args : tuple
        Additional positional arguments.
    **kwargs : dict
        Additional keyword arguments.
            newoptuna_search_kwargs: dict
                arguments for OptunaSearchCV, e.g., estimator, CV, etc.
               see  https://optuna.readthedocs.io/en/stable/reference/generated/optuna.integration.OptunaSearchCV.html
            fit_newoptuna_kwargs : dict
                fit params
            main_newoptuna_kwargs : dict
                other parameters

    Returns
    -------
    The best estimator of the estimator optimized by OptunaSearchCV.

    Methods
    -------
    prepare_data()
        Prepare data to be consumed by GridSearchCV. Pass for OptunaSearchCV case.
    optimize()
        Optimize estimator using OptunaSearchCV engine.
    get_optimized_object()
        Get the Optuna search CV object after invoking fit.
    get_best_estimator()
        Return the best estimator if already fitted.

    Notes
    -----
    It is recommended to use available factories to create a new instance of this class.
    """

    def __init__(self, X: pd.DataFrame, y: pd.DataFrame, *args, **kwargs):
        self.newoptuna_search_kwargs = kwargs["kwargs"].get(
            "newoptuna_search_kwargs", {}
        )
        self.main_newoptuna_kwargs = kwargs["kwargs"].get("main_newoptuna_kwargs", {})
        self.fit_newoptuna_kwargs = kwargs["kwargs"].get("fit_newoptuna_kwargs", {})
        self.__newoptuna_search = None
        self.__best_estimator = None
        self.__X = X
        self.__y = y
        self.args = args

    def prepare_data(self):
        """
        Prepare data to be consumed by OptunaSearchCV.

        This method prepares the data to be used by the OptunaSearchCV optimizer.
        It is not implemented in this class and should be overridden in the derived classes.
        """
        pass

    @trackcalls
    def optimize(self):
        """
        Optimize estimator using OptunaSearchCV engine.

        This method optimizes the estimator using the OptunaSearchCV engine.
        It initializes the OptunaSearchCV with the specified keyword arguments and fits the data.
        The best estimator is stored in the '__best_estimator' attribute.

        Returns
        -------
        self : NewOptunaSearch
            Returns self after optimizing the estimator.
        """
        self.__newoptuna_search = OptunaSearchCV(**self.newoptuna_search_kwargs)
        self.__newoptuna_search.fit(
            self.__X, self.__y, *self.args, **self.fit_newoptuna_kwargs
        )
        self.__best_estimator = self.__newoptuna_search.best_estimator_

        return self

    def fit(self, X: pd.DataFrame, y: pd.DataFrame, *args, **kwargs):
        """
        Fit the estimator using OptunaSearchCV.

        This method fits the estimator using the OptunaSearchCV engine.
        If the 'refit' keyword argument is set to True, the optimal model will be refit on the entire dataset.

        Parameters
        ----------
        X : pd.DataFrame
            Input for training.
        y : pd.DataFrame
            Target or label.
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        self : NewOptunaSearch
            Returns self after fitting the estimator.
        """
        for key, value in self.newoptuna_search_kwargs.items():
            if key == "refit" and value:
                self.__newoptuna_search.fit(
                    self.__X, self.__y, *self.args, **self.fit_newoptuna_kwargs
                )
                logger.info(
                    "If refit is set to True, the optimal model will be refit on the entire dataset again!"
                )
        self.__best_estimator = self.__newoptuna_search.best_estimator_
        return self

    def get_best_estimator(self):
        """
        Get the best estimator after invoking fit on it.

        Returns
        -------
        BaseEstimator or None
            The best estimator if available, otherwise None.
        """
        if self.__best_estimator is not None:
            return self.__best_estimator
        else:
            logger.error("The best estimator is None!")

    def get_optimized_object(self):
        """
        Get the Optuna search CV object after invoking fit.

        Returns
        -------
        OptunaSearchCV or None
            The Optuna search CV object if available, otherwise None.
        """
        if self.__newoptuna_search is not None:
            return self.__newoptuna_search
        else:
            raise NotImplementedError(
                "OptunaSearchCV has not been implemented or the best estimator is None."
            )


class TuneCV(OptimizerABC):
    """
    Class Factories for initializing BestModel optimizing engines, using Tune from Ray.
        Initialize the BestModelFactories.

    Parameters
    ----------
    X : pd.DataFrame or numpy.ndarray
        Input for training.
    y : pd.DataFrame or numpy.ndarray
        Target or label.
    args : tuple
        Additional positional arguments.
    kwargs : dict
        Additional keyword arguments.

        tuner_kwargs : dict, optional
            Keyword arguments for the tuner from Ray. These arguments will be passed to the tuner during initialization.
            For more information on tuner arguments, refer to the Tune documentation: https://docs.ray.io/en/latest/tune/index.html

        main_tune_kwargs : dict, optional
            Keyword arguments for the main tuning process. These arguments will be passed to the main tuning function.
            The `cv` parameter represents the cross-validation generator or an iterable used for evaluation.
            The `scoring` parameter represents the strategy to evaluate the performance of the cross-validated model on the test set.
            It can be a string, callable, list, tuple, or dictionary. Default is None.
            The `estimator` parameter represents the estimator object used for optimization.
            For scoring check https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter
            and https://scikit-learn.org/stable/modules/model_evaluation.html#scoring.
            For CV check https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html#sklearn.model_selection.GridSearchCV
            as an example.

        fit_tune_kwargs : dict, optional
            Additional keyword arguments to be passed to the `fit()` method of the estimator during training.

    Returns
    -------
    None

    Methods
    -------
    prepare_data()
        Prepare data to be consumed by TuneCV. Pass for TuneCV.
    optimize()
        Optimize estimator using OptunaSearchCV engine.
    get_optimized_object()
        Get the Optuna search cv after invoking fit.
    get_best_estimator()
        Return the best estimator if already fitted.

    Notes
    -----
    It is recommended to use available factories to create a new instance of this class.
    The `tuner_kwargs` parameter represents the keyword arguments for the tuner from Ray.
    The `fit_tune_kwargs` parameter represents the additional keyword arguments to be passed to the `fit()` method of the estimator during training.
    The `cv` parameter in `main_tune_kwargs` represents the cross-validation strategy.
    The `scoring` parameter in `main_tune_kwargs` represents the performance evaluation strategy.
    The `estimator` parameter in `main_tune_kwargs` represents the estimator object used for optimization.

    """

    def __init__(self, X, y, *args, **kwargs):
        self.tuner_kwargs = kwargs["kwargs"].get("tuner_kwargs", {})
        self.main_tune_kwargs = kwargs["kwargs"].get("main_tune_kwargs", {})
        self.fit_tune_kwargs = kwargs["kwargs"].get("fit_tune_kwargs", {})
        self.__tune = None
        self.best_estimator = None
        self.__X = X
        self.__y = y
        self.args = args

    @property
    def tuner_kwargs(self):
        return self._tuner_kwargs

    @tuner_kwargs.setter
    def tuner_kwargs(self, value):
        self._tuner_kwargs = value

    @property
    def main_tune_kwargs(self):
        return self._main_tune_kwargs

    @main_tune_kwargs.setter
    def main_tune_kwargs(self, value):
        self._main_tune_kwargs = value

    @property
    def fit_tune_kwargs(self):
        return self._fit_tune_kwargs

    @fit_tune_kwargs.setter
    def fit_tune_kwargs(self, value):
        self._fit_tune_kwargs = value

    def prepare_data(self):
        """
        Prepare data to be consumed by OptunaSearchCV.
        """
        pass

    @trackcalls
    def optimize(self):
        """
        Optimize estimator using OptunaSearchCV engine.
        """

        def trainable(config):
            """
            Trainable function for optimization.
            """

            def objective(X, y):
                """
                Objective function for optimization.
                """
                if hasattr(self.main_tune_kwargs["estimator"], "set_params"):
                    est = self.main_tune_kwargs["estimator"].set_params(**config)
                else:
                    raise AttributeError(
                        f"{self.main_tune_kwargs['estimator']} does not have the 'set_params' attribute!"
                    )

                if self.fit_tune_kwargs:
                    # Perform cross-validation with fit parameters
                    scores = cross_val_score(
                        est,
                        X=X,
                        y=y,
                        cv=self.main_tune_kwargs["cv"],
                        fit_params=self.fit_tune_kwargs,
                        scoring=self.main_tune_kwargs["scoring"],
                    )
                else:
                    # Perform cross-validation without fit parameters
                    scores = cross_val_score(
                        est,
                        X=X,
                        y=y,
                        cv=self.main_tune_kwargs["cv"],
                        scoring=self.main_tune_kwargs["scoring"],
                    )

                return scores

            # Call the objective function with the provided X and y data
            scores = objective(self.__X, self.__y)

            # Calculate the mean score
            score = np.mean(scores)

            # Report the score to Tune
            session.report({"score": score})

        # Create a tuner with the trainable function and tuner_kwargs
        tuner = tune.Tuner(trainable, **self.tuner_kwargs)

        # Fit the tuner to optimize the estimator
        results = tuner.fit()

        # Get the best result from the optimization
        best_result = results.get_best_result()

        # Get the best trial's hyperparameters
        best_config = best_result.config

        # Set the estimator with the best hyperparameters
        est = self.main_tune_kwargs["estimator"].set_params(**best_config)

        # Store the best estimator
        self.best_estimator = est

        # Store the results from Tune
        self.__tune = results

        return self

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the model to the training data.

        If the best estimator is available, fit the data using it.

        Parameters:
        -----------
        X : pd.DataFrame or numpy 2D array
            Input for training.
        y : pd.DataFrame or numpy 1D array
            Target or label.
        args : tuple
            Additional positional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns:
        --------
        self : object
            Returns self after fitting the data using the best estimator.

        """

        if self.best_estimator is not None:
            self.best_estimator.fit(
                self.__X, self.__y, *self.args, **self.fit_tune_kwargs
            )
            return self
        else:
            raise ValueError(
                "The best estimator is not available! Call the 'optimize' method first."
            )

    def get_best_estimator(self, *args, **kwargs):
        """
        Get the best estimator after invoking fit on it.

        Returns:
        --------
        best_estimator : object
            The best estimator found during the optimization process.

        Raises:
        -------
        ValueError:
            If the best estimator is not available (i.e., it is None), indicating that the fit method has not been called or has failed.

        """

        if self.best_estimator is not None:
            return self.best_estimator
        else:
            raise ValueError(
                "The best estimator is not available! Call the 'fit' method first."
            )

    def get_optimized_object(self, *args, **kwargs):
        """
        Get the Optuna search CV  after invoking fit.
        """
        if self.__tune is not None:
            return self.__tune
        else:
            raise NotImplementedError(
                "Tune has not been implemented \
                or best_results is null"
            )


class RandomSearch(OptimizerABC):

    """
    Class Factory for initializing BestModel optimizing engines, i.e., RandomizedSearchCV.
    Initialize the RandomSearch optimizer.

    Parameters
    ----------
    X : pd.DataFrame or numpy array
        Input for training.
    y : pd.DataFrame or numpy array
        Target or label.
    args : tuple
        Additional positional arguments.
    kwargs : dict
        Additional keyword arguments.
            random_search_kwargs : dict
                All arguments for creating a study using RandomSearch. Read at [RandomizedSeachCV Documentation](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
                For example:

                - estimator : object
                    An unfitted estimator that has `fit` and `predict` methods.

                - param_distributions : dict
                    Parameters that were passed to find the best estimator using the optimization method.

                - scoring : str, callable, list, tuple or dict, default=None
                    Measurement of performance for classification and regression estimator during hyperparameter optimization while estimating the best estimator.

            main_random_kwargs : dict
                Some required arguments for creating `BaseModel`:

            fit_random_kwargs : dict
                All additional parameters to the `fit()` method of the estimator during training, except `X` and `y`.


    Returns
    -------
    None

    Methods
    -------
    prepare_data()
        Prepare data to be consumed by RandomizedSearchCV.
    optimize()
        Optimize the estimator using RandomizedSearchCV engine.
    get_optimized_object()
        Get the RandomizedSearchCV object after invoking fit.
    get_best_estimator()
        Return the best estimator if already fitted.

    Notes
    -----
    It is recommended to use available factories to create a new instance of this class.
    """

    def __init__(
        self,
        X,
        y,
        *args,
        **kwargs,
    ):
        self.random_search_kwargs = kwargs["kwargs"].get("random_search_kwargs", {})
        self.main_random_kwargs = kwargs["kwargs"].get("main_random_kwargs", {})
        self.fit_random_kwargs = kwargs["kwargs"].get("fit_random_kwargs", {})
        self.__random_search = None
        self.__best_estimator = None
        self.__X = X
        self.__y = y
        self.args = args

    @property
    def random_search_kwargs(self):
        return self._random_search_kwargs

    @random_search_kwargs.setter
    def random_search_kwargs(self, value):
        self._random_search_kwargs = value

    @property
    def main_random_kwargs(self):
        return self._main_random_kwargs

    @main_random_kwargs.setter
    def main_random_kwargs(self, value):
        self._main_random_kwargs = value

    @property
    def fit_random_kwargs(self):
        return self._fit_random_kwargs

    @fit_random_kwargs.setter
    def fit_random_kwargs(self, value):
        self._fit_random_kwargs = value

    def prepare_data(self):
        """
        Prepare data to be consumed by RandomizedSearchCV.

        This method performs any necessary preprocessing steps on the data
        before applying RandomizedSearchCV.

        Returns
        -------
        None
        """
        # Implementation specific to prepare the data
        pass

    @trackcalls
    def optimize(self):
        """
        Optimize the estimator using RandomizedSearchCV engine.

        This method performs a randomized search over a hyperparameter space
        to find the best set of hyperparameters for the estimator.

        Returns
        -------
        self : object
            Returns self after optimizing the estimator.
        """
        # Create and fit the RandomizedSearchCV object
        self.__random_search = RandomizedSearchCV(**self.random_search_kwargs)
        self.__random_search.fit(
            self.__X, self.__y, *self.args, **self.fit_random_kwargs
        )

        # Set the best estimator as the result of the random search
        self.__best_estimator = self.__random_search.best_estimator_
        return self

    def fit(self, X, y, *args, **kwargs):
        """
        Get the RandomizedSearchCV object after invoking fit.

        This method returns the RandomizedSearchCV object after it has been
        fitted with the data.

        Returns
        -------
        optimized_object : object
            The optimized RandomizedSearchCV object.
        """
        # Check if refit is set to True
        if self.random_search_kwargs.get("refit", False):
            self.__random_search.fit(
                self.__X, self.__y, *self.args, **self.fit_random_kwargs
            )
            logger.info(
                "If refit is set to True, the optimal model will be refit on the entire dataset again!"
            )

        # Set the best estimator as the result of the random search
        self.__best_estimator = self.__random_search.best_estimator_

        return self

    def get_best_estimator(self, *args, **kwargs):
        """
        Return the best estimator if already fitted.

        This method returns the best estimator found during the optimization process
        if it has already been fitted.

        Returns
        -------
        best_estimator : object
            The best estimator found during the optimization process.
        """
        # Check if the best estimator is available
        if self.__best_estimator is not None:
            return self.__best_estimator
        else:
            logger.error("The best estimator is None !")

    def get_optimized_object(self, *args, **kwargs):
        """
        Get the RandomizedSearchCV object after invoking fit.

        This method returns the RandomizedSearchCV object after it has been
        fitted with the data.

        Parameters
        ----------
        args : tuple
            Additional positional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        optimized_object : object
            The optimized RandomizedSearchCV object.

        Raises
        ------
        NotImplementedError
            If the RandomizedSearchCV object is not available (i.e., it is None), indicating
            that the random search has not been implemented or the best estimator is null.
        """
        if self.__random_search is not None:
            return self.__random_search
        else:
            raise NotImplementedError(
                "RandomizedSearchCV has not been implemented or the best estimator is null."
            )


class TuneGridSearch(OptimizerABC):

    """
    Class Factories for initializing BestModel optimizing engines, i.e., TuneGridSearchCV.
        Initialize the TuneGridSearch optimizer.

    Parameters
    ----------
    X : pd.DataFrame
        Input for training.
    y : pd.DataFrame
        Target or label.
    args : tuple
        Additional positional arguments.
    kwargs : dict

        tunegrid_search_kwargs: dict
            All arguments for creating a study using TuneGridSeachCV. Read at [TuneGridSeachCV Documentation](https://docs.ray.io/en/latest/tune/api/sklearn.html)
            For example:
            - estimator : object
                An unfitted estimator that has `fit` and `predict` methods.

            - param_grid : dict
                Parameters that were passed to find the best estimator using the optimization method.

            - scoring : str
                Measurement of performance for classification and regression estimator during hyperparameter optimization while estimating the best estimator.

        main_tunegrid_kwargs : dict
            Some required arguments for creating `BaseModel`:

        fit_tunegrid_kwargs : dict
            All additional parameters to the `fit()` method of the estimator during training, except `X` and `y`.

    Returns
    -------
    None

    Methods
    -------
    prepare_data()
        Prepare data to be consumed by TuneGridSearchCV. (Placeholder method)
    optimize()
        Optimize estimator using TuneGridSearchCV engine.
    get_optimized_object()
        Get the grid search cv after invoking fit.
    get_best_estimator()
        Return the best estimator if already fitted.

    Notes
    -----
    It is recommended to use available factories to create a new instance of this class.
    """

    def __init__(self, X, y, *args, **kwargs):
        self._tunegrid_search_kwargs = kwargs.get("kwargs", {}).get(
            "tunegrid_search_kwargs", {}
        )
        self._main_tunegrid_kwargs = kwargs.get("kwargs", {}).get(
            "main_tunegrid_kwargs", {}
        )
        self._fit_tunegrid_kwargs = kwargs.get("kwargs", {}).get(
            "fit_tunegrid_kwargs", {}
        )
        self.__tunegrid_search = None
        self.__best_estimator = None
        self.__X = X
        self.__y = y
        self.args = args

    @property
    def tunegrid_search_kwargs(self):
        """
        Get the TuneGridSearchCV kwargs.

        Returns
        -------
        dict
            The TuneGridSearchCV kwargs.
        """
        return self._tunegrid_search_kwargs

    @tunegrid_search_kwargs.setter
    def tunegrid_search_kwargs(self, value):
        """
        Set the TuneGridSearchCV kwargs.

        Parameters
        ----------
        value : dict
            The TuneGridSearchCV kwargs.

        Returns
        -------
        None
        """
        self._tunegrid_search_kwargs = value

    @property
    def main_tunegrid_kwargs(self):
        """
        Get the main TuneGridSearchCV kwargs.

        Returns
        -------
        dict
            The main TuneGridSearchCV kwargs.
        """
        return self._main_tunegrid_kwargs

    @main_tunegrid_kwargs.setter
    def main_tunegrid_kwargs(self, value):
        """
        Set the main TuneGridSearchCV kwargs.

        Parameters
        ----------
        value : dict
            The main TuneGridSearchCV kwargs.

        Returns
        -------
        None
        """
        self._main_tunegrid_kwargs = value

    @property
    def fit_tunegrid_kwargs(self):
        """
        Get the fit TuneGridSearchCV kwargs.

        Returns
        -------
        dict
            The fit TuneGridSearchCV kwargs.
        """
        return self._fit_tunegrid_kwargs

    @fit_tunegrid_kwargs.setter
    def fit_tunegrid_kwargs(self, value):
        """
        Set the fit TuneGridSearchCV kwargs.

        Parameters
        ----------
        value : dict
            The fit TuneGridSearchCV kwargs.

        Returns
        -------
        None
        """
        self._fit_tunegrid_kwargs = value

    def prepare_data(self):
        """
        Prepare data to be consumed by TuneGridSearchCV.
        """
        pass

    @trackcalls
    def optimize(self):
        """
        Optimize the estimator using TuneGridSearchCV engine.

        This method optimizes the estimator using TuneGridSearchCV.
        It performs an exhaustive search over a specified parameter grid
        to find the best set of hyperparameters for the estimator.

        Returns
        -------
        self : object
            Returns self after optimizing the estimator.
        """
        self.__tunegrid_search = TuneGridSearchCV(**self.tunegrid_search_kwargs)
        self.__tunegrid_search.fit(
            self.__X, self.__y, *self.args, **self.fit_tunegrid_kwargs
        )
        self.__best_estimator = self.__tunegrid_search.best_estimator_

        return self

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the TuneGridSearchCV object with the given data.

        This method fits the TuneGridSearchCV object with the provided data.
        It can also perform refitting if the 'refit' parameter is set to True.

        Parameters
        ----------
        X : pd.DataFrame
            Input for training.
        y : pd.DataFrame
            Target or label.
        args : tuple
            Additional positional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        self : object
            Returns self after fitting the TuneGridSearchCV object.

        Notes
        -----
        If refit is set to True, the optimal model will be refit on the entire dataset again.
        """
        for key, value in self.tunegrid_search_kwargs.items():
            if key == "refit" and value:
                self.__tunegrid_search.fit(
                    self.__X, self.__y, *self.args, **self.fit_tunegrid_kwargs
                )
                logger.info(
                    "If refit is set to True, the optimal model will be refit on the entire dataset again!"
                )
        self.__best_estimator = self.__tunegrid_search.best_estimator_
        return self

    def get_best_estimator(self, *args, **kwargs):
        """
        Return the best estimator if already fitted.

        This method returns the best estimator found during the optimization process
        if it has already been fitted.

        Parameters
        ----------
        args : tuple
            Additional positional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        best_estimator : object
            The best estimator found during the optimization process.

        Notes
        -----
        If the best estimator is None, an error message is logged.
        """
        if self.__best_estimator is not None:
            return self.__best_estimator
        else:
            logger.error("The best estimator is None!")

    def get_optimized_object(self, *args, **kwargs):
        """
        Get the TuneGridSearchCV object after invoking fit.

        This method returns the TuneGridSearchCV object after it has been
        fitted with the data.

        Parameters
        ----------
        args : tuple
            Additional positional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        optimized_object : object
            The optimized TuneGridSearchCV object.

        Raises
        ------
        NotImplementedError
            If the TuneGridSearchCV object is not available (i.e., it is None), indicating
            that the grid search has not been implemented or the best estimator is null.
        """
        if self.__tunegrid_search is not None:
            return self.__tunegrid_search
        else:
            raise NotImplementedError(
                "TuneGridSearchCV has not been implemented or the best estimator is null."
            )


class TuneSearch(OptimizerABC):

    """
    Class for initializing BestModel optimizing engines, i.e., TuneSearchCV.

    This class provides factories for initializing the BestModel optimizing engine, TuneSearchCV.
    It allows for preparing data, optimizing an estimator, and accessing the optimized object and the best estimator.

    Parameters
    ----------
    X : pd.DataFrame
        Input for training.
    y : pd.DataFrame
        Target or label.
    *args : tuple
        Additional positional arguments.
    **kwargs : dict

            tune_search_kwargs: dict
                All arguments for creating a study using TuneSearchCV. Read at [TuneSearchCV Documentation](https://docs.ray.io/en/latest/tune/api/sklearn.html)

                For example:
                - estimator : object
                    An unfitted estimator that has `fit` and `predict` methods.

                - param_distributions : dict
                    Parameters that were passed to find the best estimator using the optimization method.

                - scoring : str
                    Measurement of performance for classification and regression estimator during hyperparameter optimization while estimating the best estimator.

            main_tune_kwargs : dict
                Some required arguments for creating `BaseModel`:

            fit_tune_kwargs : dict
                All additional parameters to the `fit()` method of the estimator during training, except `X` and `y`.

    Returns
    -------
    The best estimator of the estimator optimized by TuneSearchCV.

    Methods
    -------
    prepare_data()
        Prepare data to be consumed by TuneSearchCV.
    optimize()
        Optimize estimator using TuneSearchCV engine.
    get_optimized_object()
        Get the grid search CV object after invoking fit.
    get_best_estimator()
        Return the best estimator if already fitted.

    Notes
    -----
    It is recommended to use available factories to create a new instance of this class.
    """

    def __init__(self, X: pd.DataFrame, y: pd.DataFrame, *args, **kwargs):
        self.tune_search_kwargs = kwargs["kwargs"].get("tune_search_kwargs", {})
        self.main_tune_kwargs = kwargs["kwargs"].get("main_tune_kwargs", {})
        self.fit_tune_kwargs = kwargs["kwargs"].get("fit_tune_kwargs", {})
        self.__tune_search = None
        self.__best_estimator = None
        self.__X = X
        self.__y = y
        self.args = args

    def prepare_data(self):
        """
        Prepare data to be consumed by TuneSearchCV.

        This method prepares the data to be used by the TuneSearchCV optimizer.
        It is not implemented in this class and should be overridden in the derived classes.
        """
        pass

    @trackcalls
    def optimize(self):
        """
        Optimize estimator using TuneSearchCV engine.

        This method optimizes the estimator using the TuneSearchCV engine.
        It initializes the TuneSearchCV with the specified keyword arguments and fits the data.
        The best estimator is stored in the '__best_estimator' attribute.

        Returns
        -------
        self : TuneSearch
            Returns self after optimizing the estimator.
        """
        self.__tune_search = TuneSearchCV(**self.tune_search_kwargs)
        self.__tune_search.fit(self.__X, self.__y, *self.args, **self.fit_tune_kwargs)
        self.__best_estimator = self.__tune_search.best_estimator_

        return self

    def fit(self, X: pd.DataFrame, y: pd.DataFrame, *args, **kwargs):
        """
        Fit the estimator using TuneSearchCV.

        This method fits the estimator using the TuneSearchCV engine.
        If the 'refit' keyword argument is set to True, the optimal model will be refit on the entire dataset.

        Parameters
        ----------
        X : pd.DataFrame
            Input for training.
        y : pd.DataFrame
            Target or label.
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        self : TuneSearch
            Returns self after fitting the estimator.
        """
        for key, value in self.tune_search_kwargs.items():
            if key == "refit" and value:
                self.__tune_search.fit(
                    self.__X, self.__y, *self.args, **self.fit_tune_kwargs
                )
                logger.info(
                    "If refit is set to True, the optimal model will be refit on the entire dataset again!"
                )
        self.__best_estimator = self.__tune_search.best_estimator_
        return self

    def get_best_estimator(self):
        """
        Get the best estimator after invoking fit on it.

        Returns
        -------
        BaseEstimator or None
            The best estimator if available, otherwise None.
        """
        if self.__best_estimator is not None:
            return self.__best_estimator
        else:
            logger.error("The best estimator is None!")

    def get_optimized_object(self):
        """
        Get the TuneSearchCV object after invoking fit.

        Returns
        -------
        TuneSearchCV or None
            The TuneSearchCV object if available, otherwise None.
        """
        if self.__tune_search is not None:
            return self.__tune_search
        else:
            raise NotImplementedError(
                "TuneSearchCV has not been implemented or the best estimator is None."
            )



from abc import ABCMeta
from pickletools import optimize

from sklearn.base import BaseEstimator

from lohrasb import logger
from lohrasb.abstracts.estimators import AbstractEstimator
from lohrasb.base_classes.optimizer_bases import (
    GridSearch,
    NewOptunaSearch,
    OptunaSearch,
    RandomSearch,
    TuneCV,
    TuneGridSearch,
    TuneSearch,
)


class OptunaBestEstimator(AbstractEstimator):
    """
    BestModel estimation using Optuna optimization.

    This class provides functionality for estimating the best model using Optuna optimization.
    It includes methods for fitting the estimator, making predictions, and accessing the best estimator and optimized object.
    """

    def __init__(self, *args, **kwargs):
        self.best_estimator = None
        self.optimized_object = None
        self.args = args
        self.kwargs = kwargs

    def optimize(self, X, y, *args, **kwargs):
        """
        Fit the estimator using the best parameters extracted from optimization methods.
        """
        self.optimized_object = (
            OptunaSearch(X, y, *self.args, **self.kwargs).prepare_data().optimize()
        )

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the estimator using the best parameters extracted from optimization methods.
        """
        self.optimize(X, y, *args, **kwargs)
        self.optimized_object.fit(X, y, *args, **kwargs)
        self.best_estimator = self.optimized_object.get_best_estimator()

    def get_optimized_object(self):
        """
        Get the Optuna trial object after optimization.
        """
        return self.optimized_object.get_optimized_object()

    def predict(self, X):
        """
        Predict using the best estimator model.
        """
        return self.best_estimator.predict(X)

    def get_best_estimator(self):
        """
        Return the best estimator if the model is already fitted.
        """
        return self.best_estimator

    def predict_proba(self, X):
        """
        Predict class probabilities using the best estimator model.
        """
        try:
            return self.best_estimator.predict_proba(X)
        except AttributeError as e:
            raise ValueError(
                "The selected estimator does not have the predict_proba method."
            ) from e


class GridBestEstimator(AbstractEstimator):
    """
    BestModel estimation using GridSearchCV optimization.

    This class provides functionality for estimating the best model using GridSearchCV optimization.
    It includes methods for fitting the estimator, making predictions, and accessing the best estimator and optimized object.
    """

    def __init__(self, *args, **kwargs):
        self.best_estimator = None
        self.optimized_object = None
        self.args = args
        self.kwargs = kwargs

    def optimize(self, X, y, *args, **kwargs):
        """
        Fit the estimator using the best parameters extracted from optimization methods.
        """
        self.optimized_object = GridSearch(X, y, *self.args, **self.kwargs).optimize()

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the estimator using the best parameters extracted from optimization methods.
        """
        self.optimize(X, y, *args, **kwargs)
        self.optimized_object.fit(X, y, *args, **kwargs)
        self.best_estimator = self.optimized_object.get_best_estimator()

    def get_optimized_object(self):
        """
        Get the GridSearchCV object after optimization.
        """
        return self.optimized_object.get_optimized_object()

    def predict(self, X):
        """
        Predict using the best estimator model.
        """
        return self.best_estimator.predict(X)

    def get_best_estimator(self):
        """
        Return the best estimator if the model is already fitted.
        """
        return self.best_estimator

    def predict_proba(self, X):
        """
        Predict class probabilities using the best estimator model.
        """
        try:
            return self.best_estimator.predict_proba(X)
        except AttributeError as e:
            raise ValueError(
                "The selected estimator does not have the predict_proba method."
            ) from e


class NewOptunaBestEstimator(AbstractEstimator):
    """
    BestModel estimation using OptunaSearchCV optimization.

    This class provides functionality for estimating the best model using OptunaSearchCV optimization.
    It includes methods for fitting the estimator, making predictions, and accessing the best estimator and optimized object.
    """

    def __init__(self, *args, **kwargs):
        self.best_estimator = None
        self.optimized_object = None
        self.args = args
        self.kwargs = kwargs

    def optimize(self, X, y, *args, **kwargs):
        """
        Fit the feature selection estimator using the best parameters extracted from optimization methods.
        """
        self.optimized_object = NewOptunaSearch(
            X, y, *self.args, **self.kwargs
        ).optimize()

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the feature selection estimator using the best parameters extracted from optimization methods.
        """
        self.optimize(X, y, *args, **kwargs)
        self.optimized_object.fit(X, y, *args, **kwargs)
        self.best_estimator = self.optimized_object.get_best_estimator()

    def get_optimized_object(self):
        """
        Get the OptunaSearchCV object after optimization.
        """
        return self.optimized_object.get_optimized_object()

    def predict(self, X):
        """
        Predict using the best estimator model.
        """
        return self.best_estimator.predict(X)

    def get_best_estimator(self):
        """
        Return the best estimator if the model is already fitted.
        """
        return self.best_estimator

    def predict_proba(self, X):
        """
        Predict class probabilities using the best estimator model.
        """
        try:
            return self.best_estimator.predict_proba(X)
        except AttributeError as e:
            raise ValueError(
                "The selected estimator does not have the predict_proba method."
            ) from e


class RandomBestEstimator(AbstractEstimator):
    """
    BestModel estimation using RandomizedSearchCV optimization.

    This class provides functionality for estimating the best model using RandomizedSearchCV optimization.
    It includes methods for fitting the estimator, making predictions, and accessing the best estimator and optimized object.
    """

    def __init__(self, *args, **kwargs):
        self.best_estimator = None
        self.optimized_object = None
        self.args = args
        self.kwargs = kwargs

    def optimize(self, X, y, *args, **kwargs):
        """
        Fit the feature selection estimator using the best parameters extracted from optimization methods.
        """
        self.optimized_object = RandomSearch(X, y, *self.args, **self.kwargs).optimize()

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the feature selection estimator using the best parameters extracted from optimization methods.
        """
        self.optimize(X, y, *args, **kwargs)
        self.optimized_object.fit(X, y, *args, **kwargs)
        self.best_estimator = self.optimized_object.get_best_estimator()

    def get_optimized_object(self):
        """
        Get the RandomizedSearchCV object after optimization.
        """
        return self.optimized_object.get_optimized_object()

    def predict(self, X):
        """
        Predict using the best estimator model.
        """
        return self.best_estimator.predict(X)

    def get_best_estimator(self):
        """
        Return the best estimator if the model is already fitted.
        """
        return self.best_estimator

    def predict_proba(self, X):
        """
        Predict class probabilities using the best estimator model.
        """
        try:
            return self.best_estimator.predict_proba(X)
        except AttributeError as e:
            raise ValueError(
                "The selected estimator does not have the predict_proba method."
            ) from e


class TuneBestEstimator(AbstractEstimator):

    """
    BestModel estimation using Ray Tune optimization.

    This class provides functionality for estimating the best model using Ray Tune optimization.
    It includes methods for fitting the estimator, making predictions, and accessing the best estimator and optimized object.

    Parameters
    ----------
    estimator : object
        An unfitted estimator that has fit and predict methods.
    *args : tuple
        Additional positional arguments.
    **kwargs : dict
        Additional keyword arguments.

    Methods
    -------
    fit(X, y, *args, **kwargs)
        Fit the feature selection estimator using the best parameters extracted from optimization methods.
    predict(X)
        Predict using the best estimator model.
    predict_proba(X)
        Predict class probabilities using the best estimator model.
    get_best_estimator()
        Return the best estimator if already fitted.
    get_optimized_object()
        Get the TuneCV object after optimization.

    Notes
    -----
    It is recommended to use available factories to create a new instance of this class.
    """

    def __init__(self, *args, **kwargs):
        self.best_estimator = None
        self.optimized_object = None
        self.args = args
        self.kwargs = kwargs

    def optimize(self, X, y, *args, **kwargs):
        """
        Fit the feature selection estimator using the best parameters extracted from optimization methods.

        Parameters
        ----------
        X : array-like or pandas DataFrame
            Training data. Must fulfill input requirements of the feature selection step of the pipeline.
        y : array-like or pandas DataFrame or pandas Series
            Training targets. Must fulfill label requirements of the feature selection step of the pipeline.
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Additional keyword arguments.
        """
        self.optimized_object = TuneCV(X, y, *self.args, **self.kwargs).optimize()

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the feature selection estimator using the best parameters extracted from optimization methods.

        Parameters
        ----------
        X : array-like or pandas DataFrame
            Training data. Must fulfill input requirements of the feature selection step of the pipeline.
        y : array-like or pandas DataFrame or pandas Series
            Training targets. Must fulfill label requirements of the feature selection step of the pipeline.
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Additional keyword arguments.
        """
        self.optimize(X, y, *args, **kwargs)
        self.optimized_object.fit(X, y, *args, **kwargs)
        self.best_estimator = self.optimized_object.get_best_estimator()

    def get_optimized_object(self):
        """
        Get the TuneCV object after optimization.

        Returns
        -------
        TuneCV or None
            The TuneCV object if available, otherwise None.
        """
        return self.optimized_object.get_optimized_object()

    def predict(self, X):
        """
        Predict using the best estimator model.

        Parameters
        ----------
        X : array-like or pandas DataFrame
            Training data. Must fulfill input requirements of the feature selection step of the pipeline.

        Returns
        -------
        array-like
            Predicted values.
        """
        return self.best_estimator.predict(X)

    def get_best_estimator(self):
        """
        Return the best estimator if the model is already fitted.

        Returns
        -------
        BaseEstimator or None
            The best estimator if available, otherwise None.
        """
        return self.best_estimator

    def predict_proba(self, X):
        """
        Predict class probabilities using the best estimator model.

        Parameters
        ----------
        X : array-like or pandas DataFrame
            Training data. Must fulfill input requirements of the feature selection step of the pipeline.

        Returns
        -------
        array-like
            Class probabilities.

        Raises
        ------
        ValueError
            If the selected estimator does not have the predict_proba method.
        """
        try:
            return self.best_estimator.predict_proba(X)
        except AttributeError as e:
            raise ValueError(
                f"The selected estimator does not have the predict_proba method: {e}"
            )


class TuneGridBestEstimator(AbstractEstimator):
    """
    BestModel estimation using TuneGridSearchCV optimization.

    This class provides functionality for estimating the best model using TuneGridSearchCV optimization.
    It includes methods for fitting the estimator, making predictions, and accessing the best estimator and optimized object.
    """

    def __init__(self, *args, **kwargs):
        self.best_estimator = None
        self.optimized_object = None
        self.args = args
        self.kwargs = kwargs

    def optimize(self, X, y, *args, **kwargs):
        """
        Fit the estimator using the best parameters extracted from optimization methods.
        """
        self.optimized_object = TuneGridSearch(
            X, y, *self.args, **self.kwargs
        ).optimize()

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the estimator using the best parameters extracted from optimization methods.
        """
        self.optimize(X, y, *args, **kwargs)
        self.optimized_object.fit(X, y, *args, **kwargs)
        self.best_estimator = self.optimized_object.get_best_estimator()

    def get_optimized_object(self):
        """
        Get the TuneGridSearchCV object after optimization.
        """
        return self.optimized_object.get_optimized_object()

    def predict(self, X):
        """
        Predict using the best estimator model.
        """
        return self.best_estimator.predict(X)

    def get_best_estimator(self):
        """
        Return the best estimator if the model is already fitted.
        """
        return self.best_estimator

    def predict_proba(self, X):
        """
        Predict class probabilities using the best estimator model.
        """
        try:
            return self.best_estimator.predict_proba(X)
        except AttributeError as e:
            raise ValueError(
                "The selected estimator does not have the predict_proba method."
            ) from e


class TuneSearchBestEstimator(AbstractEstimator):
    """
    BestModel estimation using Tune optimization.

    This class provides functionality for estimating the best model using Tune optimization.
    It includes methods for fitting the estimator, making predictions, and accessing the best estimator and optimized object.
    """

    def __init__(self, *args, **kwargs):
        self.best_estimator = None
        self.optimized_object = None
        self.args = args
        self.kwargs = kwargs

    def optimize(self, X, y, *args, **kwargs):
        """
        Fit the estimator using the best parameters extracted from optimization methods.
        """
        self.optimized_object = TuneSearch(X, y, *self.args, **self.kwargs).optimize()

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the estimator using the best parameters extracted from optimization methods.
        """
        self.optimize(X, y, *args, **kwargs)
        self.optimized_object.fit(X, y, *args, **kwargs)
        self.best_estimator = self.optimized_object.get_best_estimator()

    def get_optimized_object(self):
        """
        Get the TuneSearchCV object after optimization.
        """
        return self.optimized_object.get_optimized_object()

    def predict(self, X):
        """
        Predict using the best estimator model.
        """
        return self.best_estimator.predict(X)

    def get_best_estimator(self):
        """
        Return the best estimator if the model is already fitted.
        """
        return self.best_estimator

    def predict_proba(self, X):
        """
        Predict class probabilities using the best estimator model.
        """
        try:
            return self.best_estimator.predict_proba(X)
        except AttributeError as e:
            raise ValueError(
                "The selected estimator does not have the predict_proba method."
            ) from e


class BaseModel(BaseEstimator, metaclass=ABCMeta):

    """
    A base class with a variety of tools for hyperparameter optimization capabilities.

    Parameters
    ----------
    kwargs : dict
        Keyword arguments for the method.

    Methods
    -------
    fit(X, y)
        Fit the feature selection estimator using the best parameters extracted from optimization methods.
    predict(X)
        Predict using the best estimator model.
    get_best_estimator()
        Return the best estimator if already fitted.

    optimize_by_gridsearchcv(cls, *args, **kwargs):
    optimize_by_optunasearchcv(cls, *args, **kwargs):
    optimize_by_randomsearchcv(cls, *args, **kwargs):
    optimize_by_tunesearchcv(cls, *args, **kwargs):
    optimize_by_optuna(cls, *args, **kwargs):
    optimize_by_tune(cls, *args, **kwargs):
    Notes
    -----
    It is recommended to use available factories to create a new instance of this class.
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @classmethod
    def optimize_by_gridsearchcv(cls, *args, **kwargs):
        """
        Optimize hyperparameters using GridSearchCV.

        Parameters
        ----------
        *args : tuple
            Additional positional arguments.

        **kwargs : dict

            grid_search_kwargs
                All arguments for creating a study using GridSearch. Read at [GridSeachCV Documentation](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html).
                For example:
                - estimator : object
                    An unfitted estimator that has `fit` and `predict` methods.

                - param_grid : dict
                    Parameters that were passed to find the best estimator using the optimization method.

                - measure_of_accuracy : str
                    Measurement of performance for classification and regression estimator during hyperparameter optimization while estimating the best estimator.

            main_grid_kwargs : dict
                Some required arguments for creating `BaseModel`:

            fit_grid_kwargs : dict
                All additional parameters to the `fit()` method of the estimator during training, except `X` and `y`.

        Returns
        -------
        GridBestEstimator
            GridBestEstimator instance.
        """
        # Ensure grid_search_kwargs, main_grid_kwargs are provided and are dictionaries
        required_kwargs = ["grid_search_kwargs"]
        for kwarg in required_kwargs:
            if kwarg not in kwargs["kwargs"]:
                raise ValueError(f"Missing required keyword argument: {kwarg}")
            if not isinstance(kwargs["kwargs"][kwarg], dict):
                raise TypeError(f"Expected a dictionary for keyword argument: {kwarg}")

        # fit_grid_kwargs should be a dictionary, if provided
        if "fit_grid_kwargs" in kwargs["kwargs"] and not isinstance(
            kwargs["kwargs"]["fit_grid_kwargs"], dict
        ):
            raise TypeError(
                "Expected a dictionary for keyword argument: fit_grid_kwargs"
            )

        # Ensure the 'estimator' key exists within the grid_search_kwargs and is an estimator
        if "estimator" not in kwargs["kwargs"]["grid_search_kwargs"]:
            raise ValueError(
                "The 'estimator' key must exist within the grid_search_kwargs."
            )
        if not hasattr(
            kwargs["kwargs"]["grid_search_kwargs"]["estimator"], "fit"
        ) or not hasattr(
            kwargs["kwargs"]["grid_search_kwargs"]["estimator"], "predict"
        ):
            raise TypeError(
                "The 'estimator' must have both 'fit' and 'predict' methods."
            )

        # Ensure the 'param_grid' key exists within the grid_search_kwargs and is a dictionary
        if "param_grid" not in kwargs["kwargs"]["grid_search_kwargs"]:
            raise ValueError(
                "The 'param_grid' key must exist within the grid_search_kwargs."
            )
        if not isinstance(kwargs["kwargs"]["grid_search_kwargs"]["param_grid"], dict):
            raise TypeError("'param_grid' must be a dictionary of parameters to tune.")

        return GridBestEstimator(**kwargs)

    @classmethod
    def optimize_by_optunasearchcv(cls, *args, **kwargs):
        """
        Optimize hyperparameters using OptunaSearchCV.

        Parameters
        ----------
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Arbitrarily keyworded arguments.
                'kwargs' : dict
                    Additional keyword arguments including:
                    - newoptuna_search_kwargs : dict
                        Arguments for configuring OptunaSearchCV, e.g., estimator, CV, etc.
                        See  https://optuna.readthedocs.io/en/stable/reference/generated/optuna.integration.OptunaSearchCV.html
                    - fit_newoptuna_kwargs : dict
                        Parameters for the `fit` method.
                    - main_newoptuna_kwargs : dict
                        Other additional parameters.

        Returns
        -------
        NewOptunaBestEstimator
            Instance of NewOptunaBestEstimator configured with provided parameters.
        """

        # Ensure kwargs is provided and is a dictionary
        if "kwargs" not in kwargs or not isinstance(kwargs["kwargs"], dict):
            raise TypeError("The 'kwargs' key must exist and must be a dictionary.")

        # Ensure newoptuna_search_kwargs, main_newoptuna_kwargs are provided and are dictionaries
        required_kwargs = ["newoptuna_search_kwargs", "main_newoptuna_kwargs"]
        for kwarg in required_kwargs:
            if kwarg not in kwargs["kwargs"]:
                raise ValueError(f"Missing required keyword argument: {kwarg}")
            if not isinstance(kwargs["kwargs"][kwarg], dict):
                raise TypeError(f"Expected a dictionary for keyword argument: {kwarg}")

        # fit_newoptuna_kwargs should be a dictionary, if provided
        if "fit_newoptuna_kwargs" in kwargs["kwargs"] and not isinstance(
            kwargs["kwargs"]["fit_newoptuna_kwargs"], dict
        ):
            raise TypeError(
                "Expected a dictionary for keyword argument: fit_newoptuna_kwargs"
            )

        # Ensure the 'estimator' key exists within the newoptuna_search_kwargs and is an estimator
        if "estimator" not in kwargs["kwargs"]["newoptuna_search_kwargs"]:
            raise ValueError(
                "The 'estimator' key must exist within the newoptuna_search_kwargs."
            )
        if not hasattr(
            kwargs["kwargs"]["newoptuna_search_kwargs"]["estimator"], "fit"
        ) or not hasattr(
            kwargs["kwargs"]["newoptuna_search_kwargs"]["estimator"], "predict"
        ):
            raise TypeError(
                "The 'estimator' within newoptuna_search_kwargs must have both 'fit' and 'predict' methods."
            )

        # Additional checks can be added as necessary...

        return NewOptunaBestEstimator(**kwargs)

    @classmethod
    def optimize_by_randomsearchcv(cls, *args, **kwargs):
        """
        Optimize hyperparameters using RandomSearchCV.

        Parameters
        ----------
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Arbitrarily keyworded arguments.
                'kwargs' : dict
                    Additional keyword arguments including:
                    - random_search_kwargs : dict
                        Arguments for configuring RandomSearchCV, e.g., estimator, param_distributions, scoring etc.
                        See  https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html
                    - fit_random_kwargs : dict
                        Parameters for the `fit` method.
                    - main_random_kwargs : dict
                        Other additional parameters.

        Returns
        -------
        RandomBestEstimator
            Instance of RandomBestEstimator configured with provided parameters.
        """

        # Ensure kwargs is provided and is a dictionary
        if "kwargs" not in kwargs or not isinstance(kwargs["kwargs"], dict):
            raise TypeError("The 'kwargs' key must exist and must be a dictionary.")

        # Ensure random_search_kwargs, main_random_kwargs are provided and are dictionaries
        required_kwargs = ["random_search_kwargs", "main_random_kwargs"]
        for kwarg in required_kwargs:
            if kwarg not in kwargs["kwargs"]:
                raise ValueError(f"Missing required keyword argument: {kwarg}")
            if not isinstance(kwargs["kwargs"][kwarg], dict):
                raise TypeError(f"Expected a dictionary for keyword argument: {kwarg}")

        # fit_random_kwargs should be a dictionary, if provided
        if "fit_random_kwargs" in kwargs["kwargs"] and not isinstance(
            kwargs["kwargs"]["fit_random_kwargs"], dict
        ):
            raise TypeError(
                "Expected a dictionary for keyword argument: fit_random_kwargs"
            )

        # Ensure the 'estimator' key exists within the random_search_kwargs and is an estimator
        if "estimator" not in kwargs["kwargs"]["random_search_kwargs"]:
            raise ValueError(
                "The 'estimator' key must exist within the random_search_kwargs."
            )
        if not hasattr(
            kwargs["kwargs"]["random_search_kwargs"]["estimator"], "fit"
        ) or not hasattr(
            kwargs["kwargs"]["random_search_kwargs"]["estimator"], "predict"
        ):
            raise TypeError(
                "The 'estimator' within random_search_kwargs must have both 'fit' and 'predict' methods."
            )

        # Additional checks can be added as necessary...

        return RandomBestEstimator(**kwargs)

    @classmethod
    def optimize_by_tunegridsearchcv(cls, *args, **kwargs):
        """
        Optimize hyperparameters using TuneGridSearchCV.

        Parameters
        ----------
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Arbitrarily keyworded arguments.
                'kwargs' : dict
                    Additional keyword arguments including:
                    - tunegrid_search_kwargs : dict
                        Arguments for configuring TuneGridSearchCV, e.g., estimator, param_grid, scoring etc.
                        See https://docs.ray.io/en/latest/tune/api/sklearn.html
                    - fit_tunegrid_kwargs : dict
                        Parameters for the `fit` method.
                    - main_tunegrid_kwargs : dict
                        Other additional parameters.

        Returns
        -------
        TuneGridBestEstimator
            Instance of TuneGridBestEstimator configured with provided parameters.
        """

        # Ensure kwargs is provided and is a dictionary
        if "kwargs" not in kwargs or not isinstance(kwargs["kwargs"], dict):
            raise TypeError("The 'kwargs' key must exist and must be a dictionary.")

        # Ensure tunegrid_search_kwargs, main_tunegrid_kwargs are provided and are dictionaries
        required_kwargs = ["tunegrid_search_kwargs", "main_tunegrid_kwargs"]
        for kwarg in required_kwargs:
            if kwarg not in kwargs["kwargs"]:
                raise ValueError(f"Missing required keyword argument: {kwarg}")
            if not isinstance(kwargs["kwargs"][kwarg], dict):
                raise TypeError(f"Expected a dictionary for keyword argument: {kwarg}")

        # fit_tune_kwargs should be a dictionary, if provided
        if "fit_tunegrid_kwargs" in kwargs["kwargs"] and not isinstance(
            kwargs["kwargs"]["fit_tunegrid_kwargs"], dict
        ):
            raise TypeError(
                "Expected a dictionary for keyword argument: fit_tunegrid_kwargs"
            )

        # Ensure the 'estimator' key exists within the tunegrid_search_kwargs and is an estimator
        if "estimator" not in kwargs["kwargs"]["tunegrid_search_kwargs"]:
            raise ValueError(
                "The 'estimator' key must exist within the tunegrid_search_kwargs."
            )
        if not hasattr(
            kwargs["kwargs"]["tunegrid_search_kwargs"]["estimator"], "fit"
        ) or not hasattr(
            kwargs["kwargs"]["tunegrid_search_kwargs"]["estimator"], "predict"
        ):
            raise TypeError(
                "The 'estimator' within tunegrid_search_kwargs must have both 'fit' and 'predict' methods."
            )

        # Additional checks can be added as necessary...

        return TuneGridBestEstimator(**kwargs)

    @classmethod
    def optimize_by_tunesearchcv(cls, *args, **kwargs):
        """
        Optimize hyperparameters using TuneSearchCV.

        Parameters
        ----------
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Arbitrarily keyworded arguments.
                'kwargs' : dict
                    Additional keyword arguments including:
                    - tune_search_kwargs : dict
                        Arguments for configuring TuneSearchCV, e.g., estimator, param_distributions, scoring etc.
                        See https://docs.ray.io/en/latest/tune/api/sklearn.html
                    - fit_tune_kwargs : dict
                        Parameters for the `fit` method.
                    - main_tune_kwargs : dict
                        Other additional parameters.

        Returns
        -------
        TuneSearchBestEstimator
            Instance of TuneSearchBestEstimator configured with provided parameters.
        """

        # Ensure kwargs is provided and is a dictionary
        if "kwargs" not in kwargs or not isinstance(kwargs["kwargs"], dict):
            raise TypeError("The 'kwargs' key must exist and must be a dictionary.")

        # Ensure tune_search_kwargs, main_tune_kwargs are provided and are dictionaries
        required_kwargs = ["tune_search_kwargs", "main_tune_kwargs"]
        for kwarg in required_kwargs:
            if kwarg not in kwargs["kwargs"]:
                raise ValueError(f"Missing required keyword argument: {kwarg}")
            if not isinstance(kwargs["kwargs"][kwarg], dict):
                raise TypeError(f"Expected a dictionary for keyword argument: {kwarg}")

        # fit_tune_kwargs should be a dictionary, if provided
        if "fit_tune_kwargs" in kwargs["kwargs"] and not isinstance(
            kwargs["kwargs"]["fit_tune_kwargs"], dict
        ):
            raise TypeError(
                "Expected a dictionary for keyword argument: fit_tune_kwargs"
            )

        # Ensure the 'estimator' key exists within the tune_search_kwargs and is an estimator
        if "estimator" not in kwargs["kwargs"]["tune_search_kwargs"]:
            raise ValueError(
                "The 'estimator' key must exist within the tune_search_kwargs."
            )
        if not hasattr(
            kwargs["kwargs"]["tune_search_kwargs"]["estimator"], "fit"
        ) or not hasattr(
            kwargs["kwargs"]["tune_search_kwargs"]["estimator"], "predict"
        ):
            raise TypeError(
                "The 'estimator' within tune_search_kwargs must have both 'fit' and 'predict' methods."
            )

        # Additional checks can be added as necessary...

        return TuneSearchBestEstimator(**kwargs)

    @classmethod
    def optimize_by_optuna(cls, *args, **kwargs):
        """
        Optimize hyperparameters using Optuna.

        Parameters
        ----------
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Arbitrarily keyworded arguments.
                'kwargs' : dict
                    Additional keyword arguments including:
                    - study_search_kwargs : dict
                        Arguments for creating a study using Optuna.
                        See https://optuna.readthedocs.io/en/stable/reference/generated/optuna.create_study.html
                    - main_optuna_kwargs : dict
                        Required arguments for `BaseModel`, e.g., estimator, estimator_params, measure_of_accuracy.
                    - optimize_kwargs : dict
                        Arguments for `optimize` method of Optuna, except `objective`.
                        See https://optuna.readthedocs.io/en/stable/reference/generated/optuna.study.Study.html#optuna.study.Study.optimize
                    - train_test_split_kwargs : dict
                        Arguments for `train_test_split` function, such as `cv`, `random_state`, etc. except the *array.
                        See https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html
                    - fit_optuna_kwargs : dict
                        Additional parameters for the `fit` method of the estimator, except `X` and `y`.

        Returns
        -------
        OptunaBestEstimator
            Instance of OptunaBestEstimator configured with provided parameters.
        """
        if "kwargs" not in kwargs or not isinstance(kwargs["kwargs"], dict):
            raise TypeError("The 'kwargs' key must exist and must be a dictionary.")

        required_kwargs = [
            "study_search_kwargs",
            "main_optuna_kwargs",
            "optimize_kwargs",
            "train_test_split_kwargs",
        ]
        for kwarg in required_kwargs:
            if kwarg not in kwargs["kwargs"]:
                raise ValueError(f"Missing required keyword argument: {kwarg}")
            if not isinstance(kwargs["kwargs"][kwarg], dict):
                raise TypeError(f"Expected a dictionary for keyword argument: {kwarg}")

        if "fit_optuna_kwargs" in kwargs["kwargs"] and not isinstance(
            kwargs["kwargs"]["fit_optuna_kwargs"], dict
        ):
            raise TypeError(
                "Expected a dictionary for keyword argument: fit_optuna_kwargs"
            )

        if "estimator" not in kwargs["kwargs"]["main_optuna_kwargs"]:
            raise ValueError(
                "The 'estimator' key must exist within the main_optuna_kwargs."
            )
        if not hasattr(
            kwargs["kwargs"]["main_optuna_kwargs"]["estimator"], "fit"
        ) or not hasattr(
            kwargs["kwargs"]["main_optuna_kwargs"]["estimator"], "predict"
        ):
            raise TypeError(
                "The 'estimator' within main_optuna_kwargs must have both 'fit' and 'predict' methods."
            )

        return OptunaBestEstimator(**kwargs)

    @classmethod
    def optimize_by_tune(cls, *args, **kwargs):
        """
        Optimize hyperparameters using Tune.

        Parameters
        ----------
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Arbitrarily keyworded arguments.
                'kwargs' : dict
                    Additional keyword arguments including:
                    - tuner_kwargs : dict
                        Arguments for the tuner from Ray.
                        See https://docs.ray.io/en/latest/tune/index.html
                    - main_tune_kwargs : dict
                        Arguments for the main tuning process including `cv`, `scoring`, `estimator`, etc.
                    - fit_tune_kwargs : dict
                        Additional parameters for the `fit` method of the estimator, except `X` and `y`.

        Returns
        -------
        TuneBestEstimator
            Instance of TuneBestEstimator configured with provided parameters.
        """
        if "kwargs" not in kwargs or not isinstance(kwargs["kwargs"], dict):
            raise TypeError("The 'kwargs' key must exist and must be a dictionary.")

        required_kwargs = ["tuner_kwargs", "main_tune_kwargs"]
        for kwarg in required_kwargs:
            if kwarg not in kwargs["kwargs"]:
                raise ValueError(f"Missing required keyword argument: {kwarg}")
            if not isinstance(kwargs["kwargs"][kwarg], dict):
                raise TypeError(f"Expected a dictionary for keyword argument: {kwarg}")

        if "fit_tune_kwargs" in kwargs["kwargs"] and not isinstance(
            kwargs["kwargs"]["fit_tune_kwargs"], dict
        ):
            raise TypeError(
                "Expected a dictionary for keyword argument: fit_tune_kwargs"
            )

        if "estimator" not in kwargs["kwargs"]["main_tune_kwargs"]:
            raise ValueError(
                "The 'estimator' key must exist within the main_tune_kwargs."
            )
        if not hasattr(
            kwargs["kwargs"]["main_tune_kwargs"]["estimator"], "fit"
        ) or not hasattr(kwargs["kwargs"]["main_tune_kwargs"]["estimator"], "predict"):
            raise TypeError(
                "The 'estimator' within main_tune_kwargs must have both 'fit' and 'predict' methods."
            )

        return TuneBestEstimator(**kwargs)

    def fit(self, X, y, *args, **kwargs):
        """
        Fit the feature selection estimator using the best parameters extracted from optimization methods.

        Parameters
        ----------
        X : Pandas DataFrame or array-like
            Training data. Must fulfill the input requirements of the feature selection step of the pipeline.
        y : Pandas DataFrame, Pandas series, or array-like
            Training targets. Must fulfill the label requirements of the feature selection step of the pipeline.
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Additional keyword arguments.
        """
        pass

    def predict(self, X):
        """
        Predict using the best estimator model.

        Parameters
        ----------
        X : Pandas DataFrame or array-like
            Input data. Must fulfill the input requirements of the feature selection step of the pipeline.
        """
        pass

    def get_best_estimator(self):
        """
        Return the best estimator if the model is already fitted.
        """
        pass



from lohrasb import logger


def _trail_param_retrive(trial, params_dict, keyword):
    """
    An internal function that returns a trial suggestion using the dictionary of estimator parameters and a specific keyword.
    Based on the keyword, it will return an Optuna.trial.suggest value.
    The returned value will be trial.suggest_int(keyword, min(params_dict[keyword]), max(params_dict[keyword])).

    Example:
        _trail_param_retrive(trial, {
            "max_depth": [2, 3],
            "min_child_weight": [0.1, 0.9],
            "gamma": [1, 9],
        }, "gamma") --> returns trial.suggest_int for gamma using [1, 9].

    Parameters:
        trial (Optuna trial): A trial is a process of evaluating an objective function.
                              For more information, visit:
                              https://optuna.readthedocs.io/en/stable/reference/generated/optuna.trial.Trial.html
        params_dict (dict): A dictionary of estimator parameters.
                            Example: {
                                "max_depth": [2, 3],
                                "min_child_weight": [0.1, 0.9],
                                "gamma": [1, 9],
                            }
        keyword (str): A keyword of the estimator's parameter.
                       Example: "gamma"
    """
    if not isinstance(params_dict, dict):
        raise TypeError("The params_dict argument must be a dictionary.")
    if not isinstance(keyword, str):
        raise TypeError("The keyword argument must be a string.")

    if keyword not in params_dict:
        raise ValueError(f"The keyword '{keyword}' is not found in the params_dict.")

    param_values = params_dict[keyword]
    if not isinstance(param_values, list):
        raise TypeError(f"The value for the keyword '{keyword}' must be a list.")

    if not param_values:
        raise ValueError(f"The list for the keyword '{keyword}' is empty.")

    if isinstance(param_values[0], str) or param_values[0] is None:
        return trial.suggest_categorical(keyword, param_values)
    if isinstance(param_values[0], int):
        if len(param_values) >= 2:
            if isinstance(param_values[1], int):
                return trial.suggest_int(keyword, min(param_values), max(param_values))
        else:
            return trial.suggest_float(keyword, min(param_values), max(param_values))
    if isinstance(param_values[0], float):
        return trial.suggest_float(keyword, min(param_values), max(param_values))


def _trail_params_retrive(trial, params_dict):
    """
    An internal function that returns trial suggestions using the dictionary of estimator parameters.

    Example:
        _trail_param_retrive(trial, {
            "eval_metric": ["auc"],
            "max_depth": [2, 3],
            "min_child_weight": [0.1, 0.9],
            "gamma": [1, 9],
            "booster": ["gbtree", "gblinear", "dart"],
        }) --> returns params where

        params = {
            "eval_metric": trial.suggest_categorical("eval_metric", ["auc"]),
            "max_depth": trial.suggest_int("max_depth", 2, 3),
            "min_child_weight": trial.suggest_float("min_child_weight", 0.1, 0.9),
            "booster": trial.suggest_categorical("booster", ["gbtree", "gblinear", "dart"]),
        }

    Parameters:
        trial (Optuna trial): A trial is a process of evaluating an objective function.
                              For more information, visit:
                              https://optuna.readthedocs.io/en/stable/reference/generated/optuna.trial.Trial.html
        params_dict (dict): A dictionary of estimator parameters.
                            Example: {
                                "eval_metric": ["auc"],
                                "max_depth": [2, 3],
                                "min_child_weight": [0.1, 0.9],
                                "gamma": [1, 9],
                                "booster": ["gbtree", "gblinear", "dart"],
                            }
    """
    if not isinstance(params_dict, dict):
        raise TypeError("The params_dict argument must be a dictionary.")

    params = {}
    for keyword in params_dict.keys():
        if not isinstance(keyword, str):
            raise TypeError("The keyword in params_dict must be a string.")

        if keyword not in params_dict:
            raise ValueError(f"The keyword '{keyword}' is not found in params_dict.")

        param_values = params_dict[keyword]
        if not isinstance(param_values, list):
            raise TypeError(f"The value for the keyword '{keyword}' must be a list.")

        if not param_values:
            raise ValueError(f"The list for the keyword '{keyword}' is empty.")

        if isinstance(param_values[0], str) or param_values[0] is None:
            params[keyword] = trial.suggest_categorical(keyword, param_values)
        elif isinstance(param_values[0], int):
            if len(param_values) >= 2:
                if isinstance(param_values[1], int):
                    params[keyword] = trial.suggest_int(
                        keyword, min(param_values), max(param_values)
                    )
            elif len(param_values) == 1:
                if isinstance(param_values[0], int):
                    params[keyword] = trial.suggest_int(
                        keyword, min(param_values), max(param_values)
                    )
            else:
                params[keyword] = trial.suggest_float(
                    keyword, min(param_values), max(param_values)
                )
        elif isinstance(param_values[0], float):
            params[keyword] = trial.suggest_float(
                keyword, min(param_values), max(param_values)
            )
        else:
            raise TypeError(
                f"The values in the list for the keyword '{keyword}' must be either strings, integers, or floats."
            )

    return params



import inspect
import os

import numpy as np
from sklearn.metrics import *

from lohrasb import logger

# for load Environment Variables
# True there will not be default args for metric


def f1_plus_tp(y_true, y_pred):
    """Return f1_score+True Positive

    This function calculates the F1 score and adds the count of true positives to it.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Calculate the F1 score using the f1_score function
    f1 = f1_score(y_true, y_pred)

    # Compute the confusion matrix using the confusion_matrix function
    # The labels argument specifies the order of the labels in the matrix
    # Here, it is set to [0, 1] indicating that label 0 corresponds to the first row, and label 1 corresponds to the second row
    _, _, _, tp = confusion_matrix(y_true, y_pred, labels=[0, 1]).ravel()

    # Add the count of true positives to the F1 score
    return f1 + tp


def f1_plus_tn(y_true, y_pred):
    """Return f1_score + True Negative

    This function calculates the F1 score and adds the count of true negatives to it.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Calculate the F1 score using the f1_score function
    f1 = f1_score(y_true, y_pred)

    # Compute the confusion matrix using the confusion_matrix function
    # The labels argument specifies the order of the labels in the matrix
    # Here, it is set to [0, 1] indicating that label 0 corresponds to the first row, and label 1 corresponds to the second row
    tn, _, _, _ = confusion_matrix(y_true, y_pred, labels=[0, 1]).ravel()

    # Add the count of true negatives to the F1 score
    return f1 + tn


def specificity(y_true, y_pred):
    """Return Specificity

    This function calculates the specificity, which is the true negative rate.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Compute the confusion matrix using the confusion_matrix function
    # The labels argument specifies the order of the labels in the matrix
    # Here, it is set to [0, 1] indicating that label 0 corresponds to the first row, and label 1 corresponds to the second row
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=[0, 1]).ravel()

    # Calculate specificity by dividing true negatives by the sum of true negatives and false positives
    return tn / (tn + fp)


def tn_score(y_true, y_pred):
    """Return True Negative (TN) Score

    This function calculates the count of true negatives.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Compute the confusion matrix using the confusion_matrix function
    # The labels argument specifies the order of the labels in the matrix
    # Here, it is set to [0, 1] indicating that label 0 corresponds to the first row, and label 1 corresponds to the second row
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=[0, 1]).ravel()

    # Return the count of true negatives
    return tn


def tn(y_true, y_pred):
    """Return True Negative (TN)

    This function calculates the count of true negatives.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Compute the confusion matrix using the confusion_matrix function
    # The labels argument specifies the order of the labels in the matrix
    # Here, it is set to [0, 1] indicating that label 0 corresponds to the first row, and label 1 corresponds to the second row
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=[0, 1]).ravel()

    # Return the count of true negatives
    return tn


def tp_score(y_true, y_pred):
    """Return True Positive (TP) Score

    This function calculates the count of true positives.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Compute the confusion matrix using the confusion_matrix function
    # The labels argument specifies the order of the labels in the matrix
    # Here, it is set to [0, 1] indicating that label 0 corresponds to the first row, and label 1 corresponds to the second row
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=[0, 1]).ravel()

    # Return the count of true positives
    return tp


def tp(y_true, y_pred):
    """Return True Positive (TP)

    This function calculates the count of true positives.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Compute the confusion matrix using the confusion_matrix function
    # The labels argument specifies the order of the labels in the matrix
    # Here, it is set to [0, 1] indicating that label 0 corresponds to the first row, and label 1 corresponds to the second row
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=[0, 1]).ravel()

    # Return the count of true positives
    return tp


def roc_plus_f1(y_true, y_pred):
    """Return ROC + F1 Score

    This function calculates the sum of the F1 score and the ROC curve.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Calculate the F1 score using the f1_score function
    f1 = f1_score(y_true, y_pred)

    # Compute the ROC curve using the roc_curve function
    roc = roc_curve(y_true, y_pred)

    # Return the sum of the F1 score and the ROC curve
    return np.sum(f1 + roc)


def auc_plus_f1(y_true, y_pred):
    """Return AUC + F1 Score

    This function calculates the sum of the F1 score and the Area Under the ROC Curve (AUC).

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Calculate the F1 score using the f1_score function
    f1 = f1_score(y_true, y_pred)

    # Compute the false positive rate (fpr), true positive rate (tpr), and thresholds for the ROC curve using the roc_curve function
    fpr, tpr, thresholds = roc_curve(y_true, y_pred)

    # Calculate the AUC using the auc function
    auc_result = auc(fpr, tpr)

    # Return the sum of the F1 score and the AUC
    return f1 + auc_result


def det_curve_ret(y_true, y_pred):
    """Return DET Curve

    This function calculates the sum of the false positive rates (fpr) and false negative rates (fnr) for the Detection Error Tradeoff (DET) curve.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Compute the false positive rate (fpr), false negative rate (fnr), and thresholds for the DET curve using the det_curve function
    fpr, fnr, thresholds = det_curve(y_true, y_pred)

    # Return the sum of the false positive rates (fpr) and false negative rates (fnr)
    return np.sum(fpr) + np.sum(fnr)


def precision_recall_curve_ret(y_true, y_pred):
    """Return Precision-Recall Curve

    This function calculates the sum of the precision and recall values for the Precision-Recall curve.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Compute the precision, recall, and thresholds for the Precision-Recall curve using the precision_recall_curve function
    precision, recall, thresholds = precision_recall_curve(y_true, y_pred)

    # Return the sum of the precision and recall values
    return np.sum(precision) + np.sum(recall)


def precision_recall_fscore_support_ret(y_true, y_pred):
    """Return Precision, Recall, F1-Score, and Support

    This function calculates the sum of the values returned by the precision_recall_fscore_support function.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Compute the precision, recall, F1-score, and support using the precision_recall_fscore_support function
    output = precision_recall_fscore_support(y_true, y_pred)

    # Return the sum of the values returned by precision_recall_fscore_support
    return np.sum(output)


def roc_curve_ret(y_true, y_pred):
    """Return ROC Curve

    This function calculates the difference between the sum of the true positive rates (tpr) and the sum of the false positive rates (fpr) for the ROC curve.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        The true values.
    y_pred : Pandas DataFrame or Pandas Series
        The predicted values.
    """

    # Compute the false positive rate (fpr), true positive rate (tpr), and thresholds for the ROC curve using the roc_curve function
    fpr, tpr, thresholds = roc_curve(y_true, y_pred)

    # Return the difference between the sum of the true positive rates (tpr) and the sum of the false positive rates (fpr)
    return np.sum(tpr) - np.sum(fpr)


class CalcMetrics:
    """
    Class for calculating metrics.

    Parameters
    ----------
    y_true : Pandas DataFrame or Pandas Series
        True values.
    y_pred : Pandas DataFrame or Pandas Series
        Predicted values.
    metric : str
        Name of the metric function.
    d : dict, optional
        A mapping dictionary that maps a metric function's string name to its representative function, by default None.
    change_default_args_of_metric : bool, optional
        Flag to change the default arguments of the metric function, by default False.
    *args, **kwargs
        Additional arguments to be passed to the metric function.

    """

    def __init__(
        self,
        y_true,
        y_pred,
        metric,
        d={
            "accuracy_score": accuracy_score,  # normal
            "auc": auc,  # normal
            "precision_recall_curve": precision_recall_curve,  # normal
            "balanced_accuracy_score": balanced_accuracy_score,  # normal
            "cohen_kappa_score": cohen_kappa_score,  # normal
            "dcg_score": dcg_score,  # normal
            "det_curve": det_curve_ret,  # normal minimize
            "f1_score": f1_score,  # normal
            "fbeta_score": fbeta_score,  # normal
            "hamming_loss": hamming_loss,  # normal minimize
            "jaccard_score": jaccard_score,  # normal
            "matthews_corrcoef": matthews_corrcoef,  # normal
            "ndcg_score": ndcg_score,  # normal
            "precision_score": precision_score,  # normal
            "recall_score": recall_score,  # normal
            "recall": recall_score,  # normal
            "roc_auc_score": roc_auc_score,  # normal
            "roc_curve": roc_curve_ret,  # normal
            "top_k_accuracy_score": top_k_accuracy_score,  # normal
            "zero_one_loss": zero_one_loss,  # normal minimize
            # customs
            "tn": tn,  # custom
            "tp": tp,  # custom
            "tn_score": tn_score,  # custom
            "tp_score": tp_score,  # custom
            "f1_plus_tp": f1_plus_tp,  # custom
            "f1_plus_tn": f1_plus_tn,  # custom
            "specificity": specificity,  # custom
            "roc_plus_f1": roc_plus_f1,  # custom
            "auc_plus_f1": auc_plus_f1,  # custom
            "precision_recall_curve_ret": precision_recall_curve_ret,  # custom
            "precision_recall_fscore_support": precision_recall_fscore_support_ret,  # custom
            # regression
            "explained_variance_score": explained_variance_score,
            "max_error": max_error,
            "mean_absolute_error": mean_absolute_error,
            "mean_squared_log_error": mean_squared_log_error,
            "mean_absolute_percentage_error": mean_absolute_percentage_error,
            "median_absolute_error": median_absolute_error,
            "r2_score": r2_score,
            "mean_poisson_deviance": mean_poisson_deviance,
            "mean_gamma_deviance": mean_gamma_deviance,
            "mean_tweedie_deviance": mean_tweedie_deviance,
        },
        change_default_args_of_metric=False,
        *args,
        **kwargs,
    ):
        self.y_true = y_true
        self.y_pred = y_pred
        self.metric = metric
        self.d = d
        self.change_default_args_of_metric = change_default_args_of_metric
        self.args = args
        self.kwargs = kwargs

    def resolve_name(self):
        """
        Resolve the metric function based on the given metric name.
        """
        if self.metric in self.d:
            return self.d[self.metric]

    def get_func_args(self):
        """
        Get the arguments of the metric function.
        """
        args_of_func = inspect.signature(self.resolve_name()).parameters
        return args_of_func

    def get_func_default_args(self):
        """
        Get the default arguments of the metric function.
        """
        args_of_func = self.get_func_args()
        d = {}
        for name, value in args_of_func.items():
            value_str = str(value)
            if "=" in value_str and value not in [
                "y_true",
                "y_score",
                "y_pred",
                "y_prob",
            ]:
                d[name] = value
        func_default_args = d
        return func_default_args

    def get_transformed_default_args(self):
        """
        Get the transformed default arguments of the metric function.
        """
        func_default_args = self.get_func_default_args()
        d = {}
        for name, value in func_default_args.items():
            value_str = str(value)
            if "None" in value_str:
                d[name] = None
            elif "True" in value_str:
                d[name] = True
            elif "False" in value_str:
                d[name] = False
            elif "'" in value_str and "=" in value_str:
                start = end = "'"
                d[name] = value_str.split(start)[1].split(end)[0]
            elif "'" not in value_str and "=" in value_str:
                start = end = "="
                d[name] = value_str.split(start)[1].split(end)[0]
            for k, v in func_default_args.items():
                try:
                    start = end = "="
                    str_var = str(v).split(start)[1].split(end)[0]
                    float_str = float(str_var)
                    d[k] = float_str
                except ValueError:
                    logger.warning(f"Warning! {k} is not used in {self.metric}")
                try:
                    start = end = "="
                    str_var = str(v).split(start)[1].split(end)[0]
                    int_str = int(str_var)
                    d[k] = int_str
                except ValueError:
                    logger.warning(f"Warning! {k} is not used in {self.metric}")

        transformed_default_args = d
        return transformed_default_args

    def assign_default(self):
        """
        Provide a way for users to change the default values of the metric function's default arguments.
        """
        metric = self.resolve_name()
        transformed_default_args = self.get_transformed_default_args()
        if self.change_default_args_of_metric:
            if len(transformed_default_args) > 0:
                ans = input(
                    f"Do you want to change the default arguments for {str(metric.__name__)}? (Y/N): "
                )
                if ans.lower() in ["y", "yes"]:
                    self.change_default_args_of_metric = True
                else:
                    self.change_default_args_of_metric = False
                if self.change_default_args_of_metric:
                    for t, v in transformed_default_args.items():
                        value_input = input(
                            f"Set a value for {t} of {str(metric.__name__)}: "
                        )
                        if value_input == "":
                            transformed_default_args[t] = v
                        elif value_input.lower() in ["true", "false"]:
                            transformed_default_args[t] = value_input.lower() == "true"
                        elif value_input.lower() == "none":
                            transformed_default_args[t] = None
                        else:
                            transformed_default_args[t] = str(value_input)

                    for t, v in transformed_default_args.items():
                        if not isinstance(v, bool):
                            try:
                                float_v = float(v)
                                transformed_default_args[t] = float_v
                            except Exception as e:
                                logger.warning(
                                    f"Warning! {t} is not used in {self.metric}"
                                )
                            try:
                                int_v = int(v)
                                transformed_default_args[t] = int_v
                            except Exception as e:
                                logger.warning(
                                    f"Warning! {t} is not used in {self.metric}"
                                )

        return transformed_default_args

    def get_default_params_if_any(self):
        """
        Get the default parameters of the metric function, if any.
        """
        assign_default = self.assign_default()
        return assign_default

    def get_func_string_if_any(self):
        """
        Get the metric function as a string, if any.
        """
        metric = self.resolve_name()
        func = str(metric.__name__)
        return func

    def get_metric_func(self):
        """
        Get the metric function and its default assigned arguments.
        """
        assign_default = self.assign_default()
        metric = self.resolve_name()
        func = str(metric.__name__)
        f_str = func + "(" + "self.y_true, self.y_pred, **assign_default" + ")"
        return eval(f_str)

    def calc_make_scorer(self, metric):
        """
        Calculate the function body of a function using make_scorer.
        See: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html

        Parameters
        ----------
        metric : str
            String representation of a metric function.
        """
        metrics_list_for_maximizing = [
            "accuracy_score",
            "auc",
            "precision_recall_curve",
            "balanced_accuracy_score",
            "cohen_kappa_score",
            "dcg_score",
            "f1_score",
            "fbeta_score",
            "fbeta_score",
            "jaccard_score",
            "matthews_corrcoef",
            "ndcg_score",
            "precision_score",
            "recall_score",
            "roc_auc_score",
            "roc_curve",
            "top_k_accuracy_score",
            # customs
            "tn",
            "tn_score",
            "tp",
            "tp_score",
            "f1_plus_tp",
            "f1_plus_tn",
            "specificity",
            "roc_plus_f1",
            "auc_plus_f1",
            "precision_recall_curve",
            "precision_recall_fscore_support",
            "max_error",
            "r2_score",
        ]
        metrics_list_for_minimizing = [
            "det_curve",
            "hamming_loss",
            "zero_one_loss",
            "explained_variance_score",
            "mean_absolute_error",
            "mean_squared_log_error",
            "mean_absolute_percentage_error",
            "mean_squared_log_error",
            "median_absolute_error",
            "mean_absolute_percentage_error",
            "mean_poisson_deviance",
            "mean_gamma_deviance",
            "mean_tweedie_deviance",
        ]

        if (
            metric.__class__.__name__ == "_BaseScorer"
            or metric.__class__.__name__ == "_ProbaScorer"
            or metric.__class__.__name__ == "_PredictScorer"
            or metric.__class__.__name__ == "_ThresholdScorer"
        ):
            return metric
        elif isinstance(metric, str):
            if "(" in metric:
                metric = metric.replace("y_true", "self.y_test")
                return metric
            else:
                raise TypeError(
                    f"The selected metric {metric} is not proper. Read the examples and documentation!"
                )
        else:
            raise TypeError(
                f"The selected metric {metric} is not proper. Read the examples and documentation!"
            )

    def get_simple_metric(self, metric, y_true, y_pred, params=None):
        """
        Get the simple metric value for y_true and y_pred.

        Parameters
        ----------
        metric : str
            String representation of a metric function.
        y_true : Pandas DataFrame or Pandas Series
            True values.
        y_pred : Pandas DataFrame or Pandas Series
            Predicted values.
        params : dict, optional
            Additional parameters for the metric function, by default None.
        """
        if params is None:
            f_str = metric + "(" + "y_true, y_pred" + ")"
            return eval(f_str)
        if isinstance(params, dict) and len(params) == 0:
            f_str = metric + "(" + "y_true, y_pred" + ")"
            return eval(f_str)
        if isinstance(params, dict) and len(params) > 0:
            f_str = metric + "(" + "y_true, y_pred, **params" + ")"
            return eval(f_str)
        return None


if __name__ == "__main__":
    # Test metrics
    calcmetric = CalcMetrics(
        y_true=np.array([0, 0, 1, 0]),
        y_pred=np.array([0, 1, 1, 0]),
        metric="d2_absolute_error_score",
    )
    calcmetric.get_metric_func()
    func = calcmetric.calc_make_scorer("d2_absolute_error_score")



# Estimators
# Linear models

import optuna

# Gradient boosting frameworks
from catboost import CatBoostClassifier, CatBoostRegressor

# Imbalanced-learn ensemble
from imblearn.ensemble import BalancedRandomForestClassifier
from lightgbm import LGBMClassifier, LGBMRegressor
from ray import air, tune
from ray.air import session

# Ensemble methods
from sklearn.ensemble import (
    AdaBoostClassifier,
    AdaBoostRegressor,
    ExtraTreesClassifier,
    ExtraTreesRegressor,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import (
    ElasticNet,
    Lasso,
    LinearRegression,
    LogisticRegression,
    Ridge,
    SGDRegressor,
)

# Other classifiers and regressors
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.svm import SVC, SVR, LinearSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from xgboost import XGBClassifier, XGBRegressor

# for gridsearchcv, randomsearchcv
# Define hyperparameters for each model
lr_params = {"penalty": ["l2"], "C": [0.001, 0.01, 0.1, 1, 10, 100]}
rf_params = {"n_estimators": [50, 100, 200], "max_depth": [None, 10, 20]}
xgb_params = {
    "n_estimators": [50, 100, 200],
    "max_depth": [6, 10, 15],
    "learning_rate": [0.001, 0.01, 0.1],
}
cb_params = {
    "iterations": [50, 100, 200],
    "depth": [4, 6, 8],
    "learning_rate": [0.001, 0.01, 0.1],
}
lgbm_params = {
    "n_estimators": [50, 100, 200],
    "max_depth": [-1, 10, 20],
    "learning_rate": [0.001, 0.01, 0.1],
}
brf_params = {"n_estimators": [50, 100, 200], "max_depth": [None, 10, 20]}
mlp_params = {
    "hidden_layer_sizes": [(50,), (100,)],
    "activation": ["relu", "tanh", "logistic"],
}
gbc_params = {
    "n_estimators": [50, 100, 200],
    "learning_rate": [0.001, 0.01, 0.1],
    "max_depth": [3, 5, 10],
}
abc_params = {"n_estimators": [50, 100, 200], "learning_rate": [0.001, 0.01, 0.1]}
svc_params = {"C": [0.1, 1, 10], "gamma": [0.001, 0.01, 0.1]}
knn_params = {"n_neighbors": [3, 5, 7], "weights": ["uniform", "distance"]}
dtc_params = {"criterion": ["gini", "entropy"], "max_depth": [None, 5, 10]}
etc_params = {"n_estimators": [50, 100, 200], "max_depth": [None, 5, 10]}
gnb_params = {}

# Put models and hyperparameters into a list of tuples
estimators_params_clfs = [
    (LogisticRegression, lr_params),
    (RandomForestClassifier, rf_params),
    (XGBClassifier, xgb_params),
    (CatBoostClassifier, cb_params),
    (LGBMClassifier, lgbm_params),
    (BalancedRandomForestClassifier, brf_params),
    (MLPClassifier, mlp_params),
    (GradientBoostingClassifier, gbc_params),
    (AdaBoostClassifier, abc_params),
    #                     (SVC, svc_params),
    #                     (KNeighborsClassifier, knn_params),
    #                     (DecisionTreeClassifier, dtc_params),
    #                     (ExtraTreesClassifier, etc_params),
    #                     (GaussianNB, gnb_params)
]

# Define hyperparameters for each model
lr_params_reg = {"fit_intercept": [True, False]}
ridge_params_reg = {"alpha": [0.5, 1.0]}
lasso_params_reg = {"alpha": [0.5, 1.0]}
elastic_params_reg = {"alpha": [0.5, 1.0], "l1_ratio": [0.5, 0.7]}
xgb_params_reg = {"max_depth": [3, 5], "learning_rate": [0.01, 0.1]}
lgbm_params_reg = {"max_depth": [3, 5], "learning_rate": [0.01, 0.1]}
cb_params_reg = {"depth": [3, 5], "learning_rate": [0.01, 0.1]}
svr_params_reg = {"kernel": ["linear", "poly", "rbf", "sigmoid"], "C": [0.5, 1.0]}
lsvr_params_reg = {"C": [0.5, 1.0]}
knr_params_reg = {
    "n_neighbors": [
        5,
        7,
        10,
        15,
        20,
    ],  # Increasing the number of neighbors can help in making the model more generalized.
    "weights": [
        "uniform",
        "distance",
    ],  # The 'distance' option can give more importance to closer instances, which may help reduce overfitting.
    "algorithm": [
        "auto",
        "ball_tree",
        "kd_tree",
        "brute",
    ],  # The algorithm used to compute the nearest neighbors can sometimes have an effect on overfitting, but it generally depends more on the dataset.
    "p": [
        1,
        2,
    ],  # This corresponds to the power parameter for the Minkowski metric. 1 is for manhattan_distance and 2 for euclidean_distance.
}
dtr_params_reg = {
    "max_depth": [None, 5, 10, 15, 20],
    "min_samples_split": [1, 2, 5, 10],
    "min_samples_leaf": [1, 2, 5, 10],
    "max_features": [None, "auto", "sqrt", "log2"],
    "max_leaf_nodes": [None, 10, 20, 30, 40],
}
rfr_params_reg = {"n_estimators": [50, 100, 200], "max_depth": [None, 10, 20]}
gbr_params_reg = {
    "n_estimators": [50, 100, 200],
    "max_depth": [3, 5, 10],
    "learning_rate": [0.001, 0.01, 0.1],
}
etr_params_reg = {"n_estimators": [50, 100, 200], "max_depth": [None, 10, 20]}
abr_params_reg = {"n_estimators": [50, 100, 200], "learning_rate": [0.001, 0.01, 0.1]}
mlpr_params_reg = {"hidden_layer_sizes": [(100,)], "activation": ["relu", "tanh"]}
sgr_params_reg = {
    "loss": ["squared_loss", "huber", "epsilon_insensitive"],
    "penalty": ["l2", "l1", "elasticnet"],
    "alpha": [0.0001, 0.001, 0.01],
}

# Put models and hyperparameters into a list of tuples
estimators_params_regs = [
    (LinearRegression, lr_params_reg),
    (Ridge, ridge_params_reg),
    (Lasso, lasso_params_reg),
    (ElasticNet, elastic_params_reg),
    (XGBRegressor, xgb_params_reg),
    (LGBMRegressor, lgbm_params_reg),
    (CatBoostRegressor, cb_params_reg),
    # (SVR, svr_params_reg),
    (LinearSVR, lsvr_params_reg),
    # TODO overfit
    # (KNeighborsRegressor, knr_params_reg),
    # TODO overfit
    # (DecisionTreeRegressor, dtr_params_reg),
    (RandomForestRegressor, rfr_params_reg),
    (GradientBoostingRegressor, gbr_params_reg),
    # (ExtraTreesRegressor, etr_params_reg),
    (AdaBoostRegressor, abr_params_reg),
    # TODO underfit
    # (MLPRegressor, mlpr_params_reg),
    # (SGDRegressor, sgr_params_reg)
]


# Define hyperparameters for each model
lr_params_clf_optuna_search = {"penalty": ["l2"], "C": [0.001, 100.0]}
rf_params_clf_optuna_search = {"n_estimators": [50, 200], "max_depth": [10, 20]}
xgb_params_clf_optuna_search = {
    "n_estimators": [50, 200],
    "max_depth": [10, 15],
    "learning_rate": [0.001, 0.1],
}
cb_params_clf_optuna_search = {
    "iterations": [50, 200],
    "depth": [4, 8],
    "learning_rate": [0.001, 0.1],
}
lgbm_params_clf_optuna_search = {
    "n_estimators": [50, 200],
    "max_depth": [10, 20],
    "learning_rate": [0.001, 0.1],
}
brf_params_clf_optuna_search = {"n_estimators": [50, 200], "max_depth": [10, 20]}
mlp_params_clf_optuna_search = {
    "hidden_layer_sizes": [(50,), (100,)],
    "activation": ["relu", "tanh", "logistic"],
}
gbc_params_clf_optuna_search = {
    "n_estimators": [50, 200],
    "learning_rate": [0.001, 0.1],
    "max_depth": [3, 10],
}
abc_params_clf_optuna_search = {
    "n_estimators": [50, 200],
    "learning_rate": [0.001, 0.1],
}
svc_params_clf_optuna_search = {"C": [0.1, 10.0], "gamma": [0.001, 0.1]}
knn_params_clf_optuna_search = {
    "n_neighbors": [3, 7],
    "weights": ["uniform", "distance"],
}
dtc_params_clf_optuna_search = {"criterion": ["gini", "entropy"], "max_depth": [5, 10]}
etc_params_clf_optuna_search = {"n_estimators": [50, 200], "max_depth": [5, 10]}


# Put models and hyperparameters into a list of tuples
estimators_params_optuna_clfs = [
    (LogisticRegression, lr_params_clf_optuna_search),
    (RandomForestClassifier, rf_params_clf_optuna_search),
    (XGBClassifier, xgb_params_clf_optuna_search),
    (CatBoostClassifier, cb_params_clf_optuna_search),
    (LGBMClassifier, lgbm_params_clf_optuna_search),
    (BalancedRandomForestClassifier, brf_params_clf_optuna_search),
    (GradientBoostingClassifier, gbc_params_clf_optuna_search),
    (AdaBoostClassifier, abc_params_clf_optuna_search),
    # (SVC, svc_params_clf_optuna_search),
    # (KNeighborsClassifier, knn_params_clf_optuna_search),
    # (DecisionTreeClassifier, dtc_params_clf_optuna_search),
    # (ExtraTreesClassifier, etc_params_clf_optuna_search),
]


# Define hyperparameters for each model
lr_params_reg_optuna_search = {"fit_intercept": [True, False]}
ridge_params_reg_optuna_search = {"alpha": [0.5, 1.0]}
lasso_params_reg_optuna_search = {"alpha": [0.5, 1.0]}
elastic_params_reg_optuna_search = {"alpha": [0.5, 1.0], "l1_ratio": [0.5, 0.7]}
xgb_params_reg_optuna_search = {"max_depth": [3, 5], "learning_rate": [0.01, 0.1]}
lgbm_params_reg_optuna_search = {"max_depth": [3, 5], "learning_rate": [0.01, 0.1]}
cb_params_reg_optuna_search = {"depth": [3, 5], "learning_rate": [0.01, 0.1]}
svr_params_reg_optuna_search = {
    "kernel": ["linear", "poly", "rbf", "sigmoid"],
    "C": [0.5, 1.0],
}
lsvr_params_reg_optuna_search = {"C": [0.5, 1.0]}
knr_params_reg_optuna_search = {
    "n_neighbors": [
        5,
        7,
        10,
        15,
        20,
    ],  # Increasing the number of neighbors can help in making the model more generalized.
    "weights": [
        "uniform",
        "distance",
    ],  # The 'distance' option can give more importance to closer instances, which may help reduce overfitting.
    "algorithm": [
        "auto",
        "ball_tree",
        "kd_tree",
        "brute",
    ],  # The algorithm used to compute the nearest neighbors can sometimes have an effect on overfitting, but it generally depends more on the dataset.
    "p": [
        1,
        2,
    ],  # This corresponds to the power parameter for the Minkowski metric. 1 is for manhattan_distance and 2 for euclidean_distance.
}
dtr_params_reg_optuna_search = {
    "max_depth": [5, 10],
    "min_samples_split": [3, 5],
    "min_samples_leaf": [2, 3],
}
rfr_params_reg_optuna_search = {"n_estimators": [50, 200], "max_depth": [10, 20]}
gbr_params_reg_optuna_search = {
    "n_estimators": [50, 200],
    "max_depth": [3, 10],
    "learning_rate": [0.01, 0.1],
}
etr_params_reg_optuna_search = {"n_estimators": [50, 200], "max_depth": [10, 20]}
abr_params_reg_optuna_search = {"n_estimators": [50, 200], "learning_rate": [0.01, 0.1]}
mlpr_params_reg_optuna_search = {
    "hidden_layer_sizes": [(100,)],
    "activation": ["relu", "tanh"],
}
sgr_params_reg_optuna_search = {
    "loss": ["huber", "epsilon_insensitive"],
    "penalty": ["l2", "l1", "elasticnet"],
    "alpha": [0.001, 0.01],
}

# Put models and hyperparameters into a list of tuples
estimators_params_optuna_regs = [
    (LinearRegression, lr_params_reg_optuna_search),
    (Ridge, ridge_params_reg_optuna_search),
    (Lasso, lasso_params_reg_optuna_search),
    (ElasticNet, elastic_params_reg_optuna_search),
    (XGBRegressor, xgb_params_reg_optuna_search),
    (LGBMRegressor, lgbm_params_reg_optuna_search),
    (CatBoostRegressor, cb_params_reg_optuna_search),
    # TODO underfit
    # (SVR, svr_params_reg_optuna_search),
    # (LinearSVR, lsvr_params_reg_optuna_search),
    # TODO overfit
    # (KNeighborsRegressor, knr_params_reg_optuna_search),
    # TODO underfit
    # (DecisionTreeRegressor, dtr_params_reg_optuna_search),
    (RandomForestRegressor, rfr_params_reg_optuna_search),
    # (GradientBoostingRegressor, gbr_params_reg_optuna_search),
    # (ExtraTreesRegressor, etr_params_reg_optuna_search),
    (AdaBoostRegressor, abr_params_reg_optuna_search),
    # (SGDRegressor, sgr_params_reg_optuna_search)
]


# Define hyperparameters for each model for optunasearchcv clfs
lr_params_optunasearchcv_clf = {
    "C": optuna.distributions.FloatDistribution(0.001, 10.0)
}
rf_params_optunasearchcv_clf = {
    "n_estimators": optuna.distributions.IntDistribution(100, 200),
    "max_depth": optuna.distributions.IntDistribution(3, 20),
}
xgb_params_optunasearchcv_clf = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "max_depth": optuna.distributions.IntDistribution(3, 15),
    "learning_rate": optuna.distributions.FloatDistribution(0.01, 0.1),
}
cb_params_optunasearchcv_clf = {
    "iterations": optuna.distributions.IntDistribution(50, 200),
    "depth": optuna.distributions.IntDistribution(4, 8),
    "learning_rate": optuna.distributions.FloatDistribution(0.01, 0.1),
}
lgbm_params_optunasearchcv_clf = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "max_depth": optuna.distributions.IntDistribution(10, 20),
    "learning_rate": optuna.distributions.FloatDistribution(0.01, 0.1),
}
brf_params_optunasearchcv_clf = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "max_depth": optuna.distributions.IntDistribution(10, 20),
}
gbc_params_optunasearchcv_clf = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "learning_rate": optuna.distributions.FloatDistribution(0.001, 0.1),
    "max_depth": optuna.distributions.IntDistribution(3, 10),
}
abc_params_optunasearchcv_clf = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "learning_rate": optuna.distributions.FloatDistribution(0.001, 0.1),
}
svc_params_optunasearchcv_clf = {
    "C": optuna.distributions.FloatDistribution(0.1, 10),
    "gamma": optuna.distributions.FloatDistribution(0.001, 0.1),
}
knn_params_optunasearchcv_clf = {
    "n_neighbors": optuna.distributions.IntDistribution(3, 7),
    "weights": optuna.distributions.CategoricalDistribution(
        choices=("uniform", "distance")
    ),
}
dtc_params_optunasearchcv_clf = {
    "criterion": optuna.distributions.CategoricalDistribution(
        choices=("gini", "entropy")
    ),
    "max_depth": optuna.distributions.IntDistribution(5, 10),
}
etc_params_optunasearchcv_clf = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "max_depth": optuna.distributions.IntDistribution(5, 10),
}
gnb_params_optunasearchcv_clf = {}

# Put models and hyperparameters into a list of tuples
estimators_params_optunasearchcv_clfs = [
    (LogisticRegression, lr_params_optunasearchcv_clf),
    (RandomForestClassifier, rf_params_optunasearchcv_clf),
    (XGBClassifier, xgb_params_optunasearchcv_clf),
    (CatBoostClassifier, cb_params_optunasearchcv_clf),
    (LGBMClassifier, lgbm_params_optunasearchcv_clf),
    (BalancedRandomForestClassifier, brf_params_optunasearchcv_clf),
    # (GradientBoostingClassifier, gbc_params_optunasearchcv_clf),
    (AdaBoostClassifier, abc_params_optunasearchcv_clf),
    # (SVC, svc_params_optunasearchcv_clf),
    # (KNeighborsClassifier, knn_params_optunasearchcv_clf),
    # (DecisionTreeClassifier, dtc_params_optunasearchcv_clf),
    # (ExtraTreesClassifier, etc_params_optunasearchcv_clf),
    # (GaussianNB, gnb_params_optunasearchcv_clf)
]


# Define hyperparameters for each model for optunasearchcv regs
lr_params_optunasearchcv_reg = {
    "fit_intercept": optuna.distributions.CategoricalDistribution(choices=(True, False))
}
ridge_params_optunasearchcv_reg = {
    "alpha": optuna.distributions.FloatDistribution(0.5, 1.0)
}
lasso_params_optunasearchcv_reg = {
    "alpha": optuna.distributions.FloatDistribution(0.5, 1.0)
}
elastic_params_optunasearchcv_reg = {
    "alpha": optuna.distributions.FloatDistribution(0.5, 1.0),
    "l1_ratio": optuna.distributions.FloatDistribution(0.5, 0.7),
}
xgb_params_optunasearchcv_reg = {
    "max_depth": optuna.distributions.IntDistribution(3, 5),
    "learning_rate": optuna.distributions.FloatDistribution(0.01, 0.1),
}
lgbm_params_optunasearchcv_reg = {
    "max_depth": optuna.distributions.IntDistribution(3, 5),
    "learning_rate": optuna.distributions.FloatDistribution(0.01, 0.1),
}
cb_params_optunasearchcv_reg = {
    "depth": optuna.distributions.IntDistribution(3, 5),
    "learning_rate": optuna.distributions.FloatDistribution(0.01, 0.10),
}
svr_params_optunasearchcv_reg = {
    "kernel": optuna.distributions.CategoricalDistribution(
        choices=("linear", "poly", "rbf", "sigmoid")
    ),
    "C": optuna.distributions.FloatDistribution(0.1, 1.0),
}
lsvr_params_optunasearchcv_reg = {"C": optuna.distributions.FloatDistribution(0.5, 1.0)}
knr_params_optunasearchcv_reg = {
    "n_neighbors": optuna.distributions.IntDistribution(
        30, 50
    ),  # Increasing the number of neighbors can help in making the model more generalized.
    "weights": optuna.distributions.CategoricalDistribution(
        choices=("uniform", "distance")
    ),  # The 'distance' option can give more importance to closer instances, which may help reduce overfitting.
    "algorithm": optuna.distributions.CategoricalDistribution(
        choices=("auto", "ball_tree", "kd_tree", "brute")
    ),  # The algorithm used to compute the nearest neighbors can sometimes have an effect on overfitting, but it generally depends more on the dataset.
    "p": optuna.distributions.IntDistribution(
        1, 2
    ),  # This corresponds to the power parameter for the Minkowski metric. 1 is for manhattan_distance and 2 for euclidean_distance.
}
dtr_params_optunasearchcv_reg = {
    "max_depth": optuna.distributions.IntDistribution(3, 5),
    "min_samples_split": optuna.distributions.IntDistribution(2, 5),
    "min_samples_leaf": optuna.distributions.IntDistribution(2, 5),
    "max_features": optuna.distributions.CategoricalDistribution(
        choices=("auto", "sqrt", "log2")
    ),
    "max_leaf_nodes": optuna.distributions.IntDistribution(5, 10),
}
rfr_params_optunasearchcv_reg = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "max_depth": optuna.distributions.IntDistribution(10, 20),
}
gbr_params_optunasearchcv_reg = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "max_depth": optuna.distributions.IntDistribution(3, 10),
    "learning_rate": optuna.distributions.FloatDistribution(0.001, 0.1),
}
etr_params_optunasearchcv_reg = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "max_depth": optuna.distributions.IntDistribution(10, 20),
}
abr_params_optunasearchcv_reg = {
    "n_estimators": optuna.distributions.IntDistribution(50, 200),
    "learning_rate": optuna.distributions.FloatDistribution(0.001, 0.1),
}
sgr_params_optunasearchcv_reg = {
    "loss": optuna.distributions.CategoricalDistribution(
        choices=("squared_loss", "huber", "epsilon_insensitive")
    ),
    "penalty": optuna.distributions.CategoricalDistribution(
        choices=("l2", "l1", "elasticnet")
    ),
    "alpha": optuna.distributions.FloatDistribution(0.0001, 0.01),
}

# Put models and hyperparameters into a list of tuples
estimators_params_optunasearchcv_regs = [
    (LinearRegression, lr_params_optunasearchcv_reg),
    (Ridge, ridge_params_optunasearchcv_reg),
    (Lasso, lasso_params_optunasearchcv_reg),
    (ElasticNet, elastic_params_optunasearchcv_reg),
    (XGBRegressor, xgb_params_optunasearchcv_reg),
    (LGBMRegressor, lgbm_params_optunasearchcv_reg),
    (CatBoostRegressor, cb_params_optunasearchcv_reg),
    # TODO underfit
    # (SVR, svr_params_optunasearchcv_reg),
    # (LinearSVR, lsvr_params_optunasearchcv_reg),
    # TODO overfit
    # (KNeighborsRegressor, knr_params_optunasearchcv_reg),
    # TODO underfit
    # (DecisionTreeRegressor, dtr_params_optunasearchcv_reg),
    (RandomForestRegressor, rfr_params_optunasearchcv_reg),
    # (GradientBoostingRegressor, gbr_params_optunasearchcv_reg),
    # (ExtraTreesRegressor, etr_params_optunasearchcv_reg),
    (AdaBoostRegressor, abr_params_optunasearchcv_reg),
    # (SGDRegressor, sgr_params_optunasearchcv_reg)
]

# Define hyperparameters for tunesearch classification
lr_params_tunesearch_clfs = {"C": tune.uniform(0.001, 10.0)}
rf_params_tunesearch_clfs = {
    "n_estimators": tune.randint(100, 200),
    "max_depth": tune.randint(3, 20),
}
xgb_params_tunesearch_clfs = {
    "n_estimators": tune.randint(50, 200),
    "max_depth": tune.randint(3, 15),
    "learning_rate": tune.uniform(0.01, 0.1),
}
cb_params_tunesearch_clfs = {
    "iterations": tune.randint(50, 200),
    "depth": tune.randint(4, 8),
    "learning_rate": tune.uniform(0.01, 0.1),
}
lgbm_params_tunesearch_clfs = {
    "n_estimators": tune.randint(50, 200),
    "max_depth": tune.randint(10, 20),
    "learning_rate": tune.uniform(0.01, 0.1),
}
brf_params_tunesearch_clfs = {
    "n_estimators": tune.randint(50, 200),
    "max_depth": tune.randint(10, 20),
}
gbc_params_tunesearch_clfs = {
    "n_estimators": tune.randint(50, 200),
    "learning_rate": tune.uniform(0.001, 0.1),
    "max_depth": tune.randint(3, 10),
}
abc_params_tunesearch_clfs = {
    "n_estimators": tune.randint(50, 200),
    "learning_rate": tune.uniform(0.001, 0.1),
}
svc_params_tunesearch_clfs = {
    "C": tune.uniform(0.1, 10),
    "gamma": tune.uniform(0.001, 0.1),
}
knn_params_tunesearch_clfs = {
    "n_neighbors": tune.randint(3, 7),
    "weights": tune.choice(["uniform", "distance"]),
}
dtc_params_tunesearch_clfs = {
    "criterion": tune.choice(["gini", "entropy"]),
    "max_depth": tune.randint(5, 10),
}
etc_params_tunesearch_clfs = {
    "n_estimators": tune.randint(50, 200),
    "max_depth": tune.randint(5, 10),
}
# gnb_params_tunesearch_clfs = {}

# Put models and hyperparameters into a list of tuples
estimators_params_tunesearch_clfs = [
    (LogisticRegression, lr_params_tunesearch_clfs),
    (RandomForestClassifier, rf_params_tunesearch_clfs),
    (XGBClassifier, xgb_params_tunesearch_clfs),
    (CatBoostClassifier, cb_params_tunesearch_clfs),
    (LGBMClassifier, lgbm_params_tunesearch_clfs),
    (BalancedRandomForestClassifier, brf_params_tunesearch_clfs),
    (GradientBoostingClassifier, gbc_params_tunesearch_clfs),
    (AdaBoostClassifier, abc_params_tunesearch_clfs),
    (SVC, svc_params_tunesearch_clfs),
    # (KNeighborsClassifier, knn_params_tunesearch_clfs),
    (DecisionTreeClassifier, dtc_params_tunesearch_clfs),
    (ExtraTreesClassifier, etc_params_tunesearch_clfs),
    # (GaussianNB, gnb_params_tunesearch_clfs),
]

# Define hyperparameters for tune regression
lr_params_tunesearch_regs = {"fit_intercept": tune.choice([True, False])}
ridge_params_tunesearch_regs = {"alpha": tune.uniform(0.5, 1.0)}
lasso_params_tunesearch_regs = {"alpha": tune.uniform(0.5, 1.0)}
elastic_params_tunesearch_regs = {
    "alpha": tune.uniform(0.5, 1.0),
    "l1_ratio": tune.uniform(0.5, 0.7),
}
xgb_params_tunesearch_regs = {
    "max_depth": tune.randint(3, 5),
    "learning_rate": tune.uniform(0.01, 0.1),
}
lgbm_params_tunesearch_regs = {
    "max_depth": tune.randint(3, 5),
    "learning_rate": tune.uniform(0.01, 0.1),
}
cb_params_tunesearch_regs = {
    "depth": tune.randint(3, 5),
    "learning_rate": tune.uniform(0.01, 0.10),
}
svr_params_tunesearch_regs = {
    "kernel": tune.choice(["linear", "poly", "rbf", "sigmoid"]),
    "C": tune.uniform(0.1, 1.0),
}
lsvr_params_tunesearch_regs = {"C": tune.uniform(0.5, 1.0)}
knr_params_tunesearch_regs = {
    "n_neighbors": tune.randint(
        30, 50
    ),  # Increasing the number of neighbors can help in making the model more generalized.
    "weights": tune.choice(
        ["uniform", "distance"]
    ),  # The 'distance' option can give more importance to closer instances, which may help reduce overfitting.
    "algorithm": tune.choice(
        ["auto", "ball_tree", "kd_tree", "brute"]
    ),  # The algorithm used to compute the nearest neighbors can sometimes have an effect on overfitting, but it generally depends more on the dataset.
    "p": tune.randint(
        1, 2
    ),  # This corresponds to the power parameter for the Minkowski metric. 1 is for manhattan_distance and 2 for euclidean_distance.
}
dtr_params_tunesearch_regs = {
    "max_depth": tune.randint(3, 5),
    "min_samples_split": tune.randint(2, 5),
    "min_samples_leaf": tune.randint(2, 5),
    "max_features": tune.choice(["auto", "sqrt", "log2"]),
    "max_leaf_nodes": tune.randint(5, 10),
}
rfr_params_tunesearch_regs = {
    "n_estimators": tune.randint(50, 200),
    "max_depth": tune.randint(10, 20),
}
gbr_params_tunesearch_regs = {
    "n_estimators": tune.randint(50, 200),
    "max_depth": tune.randint(3, 10),
    "learning_rate": tune.uniform(0.001, 0.1),
}
etr_params_tunesearch_regs = {
    "n_estimators": tune.randint(50, 200),
    "max_depth": tune.randint(10, 20),
}
abr_params_tunesearch_regs = {
    "n_estimators": tune.randint(50, 200),
    "learning_rate": tune.uniform(0.001, 0.1),
}
sgr_params_tunesearch_regs = {
    "loss": tune.choice(["squared_loss", "huber", "epsilon_insensitive"]),
    "penalty": tune.choice({"l2", "l1", "elasticnet"}),
    "alpha": tune.uniform(0.0001, 0.01),
}

# Put models and hyperparameters into a list of tuples
estimators_params_tunesearch_regs = [
    (LinearRegression, lr_params_tunesearch_regs),
    (Ridge, ridge_params_tunesearch_regs),
    (Lasso, lasso_params_tunesearch_regs),
    (ElasticNet, elastic_params_tunesearch_regs),
    (XGBRegressor, xgb_params_tunesearch_regs),
    (LGBMRegressor, lgbm_params_tunesearch_regs),
    (CatBoostRegressor, cb_params_tunesearch_regs),
    # TODO underfit
    # (SVR, svr_params_tunesearch_regs),
    # (LinearSVR, lsvr_params_tunesearch_regs),
    # TODO overfit
    # (KNeighborsRegressor, knr_params_tunesearch_regs),
    # TODO underfit
    # (DecisionTreeRegressor, dtr_params_tunesearch_regs),
    (RandomForestRegressor, rfr_params_tunesearch_regs),
    # (GradientBoostingRegressor, gbr_params_tunesearch_regs),
    (ExtraTreesRegressor, etr_params_tunesearch_regs),
    (AdaBoostRegressor, abr_params_tunesearch_regs),
    # (SGDRegressor, sgr_params_tunesearch_regs),
]



__version__ = "4.2.0"

import logging
import logging.config
import os

from dotenv import load_dotenv
from ruamel.yaml import YAML

from .project_conf import ROOT_PROJECT

yaml = YAML()


PATH_TO_LOG_CONF = ROOT_PROJECT / "lohrasb" / "config.yaml"

# DEFAULT_LEVEL in production env
DEFAULT_LEVEL = logging.ERROR


def log_setup(log_cfg_path=PATH_TO_LOG_CONF):
    try:
        with open(PATH_TO_LOG_CONF, "r") as f:
            config = yaml.load(f.read())
            logging.config.dictConfig(config)
            # set up logging configuration
            return True
    except Exception as e:
        print(e)
        print(
            f"In this module, the default logging will be applied. The error is {e} which will be skipped!"
        )
        return False


if log_setup():
    # create logger
    load_dotenv()
    env = os.getenv("env")
    logger = logging.getLogger(env)

else:
    print("default logger setting is applied !")
    logging.basicConfig(level=DEFAULT_LEVEL)
    logger = logging.getLogger()



from setuptools import setup, find_packages

# Read the README.md for long description
with open("README.md", "r") as fh:
    long_description = fh.read()

# Read the requirements from requirements_prod.txt
with open("requirements_prod.txt", "r") as f:
    requirements = f.read().splitlines()

setup(
    name='lohrasb',
    version='4.2.0',
    author='drhosseinjavedani',
    author_email='h.javedani@gmail.com',
    description=("This versatile tool streamlines hyperparameter optimization in machine learning workflows."
                 "It supports a wide range of search methods, from GridSearchCV and RandomizedSearchCV"
                 "to advanced techniques like OptunaSearchCV, Ray Tune, and Scikit-Learn Tune."
                 "Designed to enhance model performance and efficiency, it's suitable for tasks of any scale."),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='',
    license='BSD-3-Clause license',
    packages=find_packages(exclude=["examples*"]),
    include_package_data=True,  # This will read MANIFEST.in
    keywords=["Auto ML", "Pipeline", "Machine learning"],
    install_requires=requirements  # Use the parsed requirements here
)
